"""
Module for fetching files via HTTP and FTP. Directly or over SSL (HTTPS) with
server certificate validation.

"""

import re
import os
import os.path
import pycurl

from pyanaconda.flags import flags as ana_flags

from org_fedora_oscap import utils

import logging
log = logging.getLogger("anaconda")


# everything else should be private
__all__ = ["fetch_data", "can_fetch_from"]

# prefixes of the URLs that need network connection
NET_URL_PREFIXES = ("http", "https", "ftp")

# prefixes of the URLs that may not need network connection
LOCAL_URL_PREFIXES = ("file",)

# TODO: needs improvements
HTTP_URL_RE_STR = r"(https?)://(.*)"
HTTP_URL_RE = re.compile(HTTP_URL_RE_STR)

FTP_URL_RE_STR = r"(ftp)://(.*)"
FTP_URL_RE = re.compile(FTP_URL_RE_STR)

FILE_URL_RE_STR = r"(file)://(.*)"
FILE_URL_RE = re.compile(FILE_URL_RE_STR)


class DataFetchError(Exception):
    """Parent class for the exception classes defined in this module."""

    pass


class CertificateValidationError(DataFetchError):
    """Class for the certificate validation related errors."""

    pass


class WrongRequestError(DataFetchError):
    """Class for the wrong combination of parameters errors."""

    pass


class UnknownURLformatError(DataFetchError):
    """Class for invalid URL cases."""

    pass


class FetchError(DataFetchError):
    """
    Class for the errors when fetching data. Usually due to I/O errors.

    """

    pass


def can_fetch_from(url):
    """
    Function telling whether the fetch_data function understands the type of
    given URL or not.

    :param url: URL
    :type url: str
    :return: whether the type of the URL is supported or not
    :rtype: str

    """
    resources = NET_URL_PREFIXES + LOCAL_URL_PREFIXES
    return any(url.startswith(prefix) for prefix in resources)


def fetch_data(url, out_file, ca_certs=None):
    """
    Fetch data from a given URL. If the URL starts with https://, ca_certs can
    be a path to PEM file with CA certificate chain to validate server
    certificate.

    :param url: URL of the data
    :type url: str
    :param out_file: path to the output file
    :type out_file: str
    :param ca_certs: path to a PEM file with CA certificate chain
    :type ca_certs: str
    :raise WrongRequestError: if a wrong combination of arguments is passed
                              (ca_certs file path given and url starting with
                              http://) or arguments don't have required format
    :raise CertificateValidationError: if server certificate validation fails
    :raise FetchError: if data fetching fails (usually due to I/O errors)

    """

    # create the directory for the out_file if it doesn't exist
    out_dir = os.path.dirname(out_file)
    utils.ensure_dir_exists(out_dir)

    if can_fetch_from(url):
        _curl_fetch(url, out_file, ca_certs)
    else:
        msg = "Cannot fetch data from '%s': unknown URL format" % url
        raise UnknownURLformatError(msg)


def _curl_fetch(url, out_file, ca_certs=None):
    """
    Function that fetches data and writes it out to the given file path. If a
    path to the file with CA certificates is given and the url starts with
    'https', the server certificate is validated.

    :param url: url of the data that has to start with 'http://' or "https://"
    :type url: str
    :param out_file: path to the output file
    :type out_file: str
    :param ca_certs: path to the file with CA certificates for server
                     certificate validation
    :type ca_certs: str
    :raise WrongRequestError: if a wrong combination of arguments is passed
                              (ca_certs file path given and url starting with
                              http://) or arguments don't have required format
    :raise CertificateValidationError: if server certificate validation fails
    :raise FetchError: if data fetching fails (usually due to I/O errors)

    """

    if url.startswith("ftp"):
        match = FTP_URL_RE.match(url)
        if not match:
            msg = "Wrong url not matching '%s'" % FTP_URL_RE_STR
            raise WrongRequestError(msg)
        else:
            protocol, path = match.groups()
            if '@' not in path:
                # no user:pass given -> use anonymous login to the FTP server
                url = protocol + "://anonymous:@" + path
    elif url.startswith("file"):
        match = FILE_URL_RE.match(url)
        if not match:
            msg = "Wrong url not matching '%s'" % FILE_URL_RE_STR
            raise WrongRequestError(msg)
    else:
        match = HTTP_URL_RE.match(url)
        if not match:
            msg = "Wrong url not matching '%s'" % HTTP_URL_RE_STR
            raise WrongRequestError(msg)

    # the first group contains the protocol, the second one the rest
    protocol = match.groups()[0]

    if not out_file:
        raise WrongRequestError("out_file cannot be an empty string")

    if ca_certs and protocol != "https":
        msg = "Cannot verify server certificate when using plain HTTP"
        raise WrongRequestError(msg)

    curl = pycurl.Curl()
    curl.setopt(pycurl.URL, url)

    if ca_certs and protocol == "https":
        # the strictest verification
        curl.setopt(pycurl.SSL_VERIFYHOST, 2)
        curl.setopt(pycurl.SSL_VERIFYPEER, 1)
        curl.setopt(pycurl.CAINFO, ca_certs)

    # may be turned off by flags (specified on command line, take precedence)
    if ana_flags.noverifyssl:
        log.warning("Disabling SSL verification due to the noverifyssl flag")
        curl.setopt(pycurl.SSL_VERIFYHOST, 0)
        curl.setopt(pycurl.SSL_VERIFYPEER, 0)

    try:
        with open(out_file, "w") as fobj:
            curl.setopt(pycurl.WRITEDATA, fobj)
            curl.perform()
    except pycurl.error as err:
        # first arg is the error code
        if err.args[0] == pycurl.E_SSL_CACERT:
            msg = "Failed to connect to server and validate its "\
                  "certificate: %s" % err
            raise CertificateValidationError(msg)
        else:
            msg = "Failed to fetch data: %s" % err
            raise FetchError(msg)



#
# Copyright (C) 2013  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Vratislav Podzimek <vpodzime@redhat.com>
#

"""Module with unit tests for the rule_handling.py module"""

import unittest
import mock

from org_fedora_oscap import rule_handling, common


class PartRulesSyntaxSupportTest(unittest.TestCase):
    """Test functionality of the PartRules' methods with syntax support."""

    def setUp(self):
        self.part_rules = rule_handling.PartRules()
        self.part_rules.ensure_mount_point("/tmp")

    # simple tests, shouldn't raise exceptions
    def getitem_test(self):
        self.part_rules["/tmp"]

    def setitem_test(self):
        rule = rule_handling.PartRule("/var/log")
        self.part_rules["/var/log"] = rule

    def len_test(self):
        self.assertEqual(len(self.part_rules), 1)

    def contains_test(self):
        self.assertTrue("/tmp" in self.part_rules)

    def delitem_test(self):
        del(self.part_rules["/tmp"])
        self.assertNotIn("/tmp", self.part_rules)


class RuleDataParsingTest(unittest.TestCase):
    """Test rule data parsing."""

    def setUp(self):
        self.rule_data = rule_handling.RuleData()

    def artificial_test(self):
        self.rule_data.new_rule("  part /tmp --mountoptions=nodev,noauto")
        self.rule_data.new_rule("part /var/log  ")
        self.rule_data.new_rule(" passwd   --minlen=14 ")
        self.rule_data.new_rule("package --add=iptables")
        self.rule_data.new_rule(" package --add=firewalld --remove=telnet")
        self.rule_data.new_rule("package --remove=rlogin --remove=sshd")
        self.rule_data.new_rule("bootloader --passwd")

        # both partitions should appear in self.rule_data._part_rules
        self.assertIn("/tmp", self.rule_data._part_rules)
        self.assertIn("/var/log", self.rule_data._part_rules)

        # mount options should be parsed
        self.assertIn("nodev",
                      self.rule_data._part_rules["/tmp"]._mount_options)
        self.assertIn("noauto",
                      self.rule_data._part_rules["/tmp"]._mount_options)

        # no mount options for /var/log
        self.assertEqual(self.rule_data._part_rules["/var/log"]._mount_options,
                         [])

        # minimal password length should be parsed and stored correctly
        self.assertEqual(self.rule_data._passwd_rules._minlen, 14)

        # packages should be parsed correctly
        self.assertIn("iptables", self.rule_data._package_rules._add_pkgs)
        self.assertIn("firewalld", self.rule_data._package_rules._add_pkgs)
        self.assertIn("telnet", self.rule_data._package_rules._remove_pkgs)
        self.assertIn("rlogin", self.rule_data._package_rules._remove_pkgs)
        self.assertIn("sshd", self.rule_data._package_rules._remove_pkgs)

        # bootloader should require password
        self.assertTrue(self.rule_data._bootloader_rules._require_password)

    def quoted_opt_values_test(self):
        self.rule_data.new_rule('part /tmp --mountoptions="nodev,noauto"')

        self.assertIn("nodev",
                      self.rule_data._part_rules["/tmp"]._mount_options)
        self.assertIn("noauto",
                      self.rule_data._part_rules["/tmp"]._mount_options)
        self.assertNotIn('"',
                         self.rule_data._part_rules["/tmp"]._mount_options)

    def real_output_test(self):
        output = """
      part /tmp

      part /tmp --mountoptions=nodev
    """
        for line in output.splitlines():
            self.rule_data.new_rule(line)

        self.assertIn("/tmp", self.rule_data._part_rules)
        self.assertIn("nodev",
                      self.rule_data._part_rules["/tmp"]._mount_options)

        # should be stripped and merged
        self.assertEqual(str(self.rule_data._part_rules),
                         "part /tmp --mountoptions=nodev")


class RuleEvaluationTest(unittest.TestCase):
    """Test if the rule evaluation works properly."""

    def setUp(self):
        self.rule_data = rule_handling.RuleData()
        self.ksdata_mock = mock.Mock()
        self.storage_mock = mock.Mock()

    def existing_part_must_exist_rules_test(self):
        for rule in ["part /tmp", "part /"]:
            self.rule_data.new_rule(rule)

        tmp_part_mock = mock.Mock()
        tmp_part_mock.format.options = "defaults"
        root_part_mock = mock.Mock()
        root_part_mock.format.options = "defaults"

        self.storage_mock.mountpoints = {"/tmp": tmp_part_mock,
                                         "/": root_part_mock,
                                         }

        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock)

        # partitions exist --> no errors, warnings or additional info
        self.assertEqual(messages, [])

        # no additional mount options specified
        self.assertEqual(tmp_part_mock.format.options, "defaults")
        self.assertEqual(root_part_mock.format.options, "defaults")

    def nonexisting_part_must_exist_test(self):
        for rule in ["part /tmp", "part /"]:
            self.rule_data.new_rule(rule)

        tmp_part_mock = mock.Mock()
        tmp_part_mock.format.options = "defaults"

        self.storage_mock.mountpoints = {"/tmp": tmp_part_mock,
                                         }

        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock)

        # / mount point missing --> one error
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0].type, common.MESSAGE_TYPE_FATAL)

        # error has to mention the mount point
        self.assertIn("/", messages[0].text)

    def add_mount_options_test(self):
        for rule in ["part /tmp --mountoptions=defaults,nodev",
                     "part / --mountoptions=noauto"]:
            self.rule_data.new_rule(rule)

        tmp_part_mock = mock.Mock()
        tmp_part_mock.format.options = "defaults"
        root_part_mock = mock.Mock()
        root_part_mock.format.options = "defaults"

        self.storage_mock.mountpoints = {"/tmp": tmp_part_mock,
                                         "/": root_part_mock,
                                         }

        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock)

        # two mount options added --> two info messages
        self.assertEqual(len(messages), 2)
        self.assertTrue(all(message.type == common.MESSAGE_TYPE_INFO
                            for message in messages))

        # newly added mount options should be mentioned in the messages
        # together with their mount points
        nodev_found = False
        noauto_found = False

        for message in messages:
            if "'nodev'" in message.text:
                self.assertIn("/tmp", message.text)
                nodev_found = True
            elif "'noauto'" in message.text:
                self.assertIn("/", message.text)
                noauto_found = True

        self.assertTrue(all([nodev_found, noauto_found]))
        self.assertEqual(self.storage_mock.mountpoints["/tmp"].format.options,
                         "defaults,nodev")
        self.assertEqual(self.storage_mock.mountpoints["/"].format.options,
                         "defaults,noauto")

    def add_mount_options_no_duplicates_test(self):
        for rule in ["part /tmp --mountoptions=defaults,nodev",
                     "part / --mountoptions=noauto"]:
            self.rule_data.new_rule(rule)

        tmp_part_mock = mock.Mock()
        tmp_part_mock.format.options = "defaults"
        root_part_mock = mock.Mock()
        root_part_mock.format.options = "defaults"

        self.storage_mock.mountpoints = {"/tmp": tmp_part_mock,
                                         "/": root_part_mock,
                                         }

        # evaluate twice, so that duplicates could possible by created
        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock)
        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock)

        # two mount options added --> two info messages
        self.assertEqual(len(messages), 2)
        self.assertTrue(all(message.type == common.MESSAGE_TYPE_INFO
                            for message in messages))

        # newly added mount options should be mentioned in the messages
        # together with their mount points
        nodev_found = False
        noauto_found = False

        for message in messages:
            if "'nodev'" in message.text:
                self.assertIn("/tmp", message.text)
                nodev_found = True
            elif "'noauto'" in message.text:
                self.assertIn("/", message.text)
                noauto_found = True

        self.assertTrue(all([nodev_found, noauto_found]))

        # no duplicates should be added (each new mount option only once)
        self.assertEqual(self.storage_mock.mountpoints["/tmp"].format.options,
                         "defaults,nodev")
        self.assertEqual(self.storage_mock.mountpoints["/"].format.options,
                         "defaults,noauto")

    def add_mount_options_report_only_test(self):
        for rule in ["part /tmp --mountoptions=nodev",
                     "part / --mountoptions=noauto"]:
            self.rule_data.new_rule(rule)

        tmp_part_mock = mock.Mock()
        tmp_part_mock.format.options = "defaults"
        root_part_mock = mock.Mock()
        root_part_mock.format.options = "defaults"

        self.storage_mock.mountpoints = {"/tmp": tmp_part_mock,
                                         "/": root_part_mock,
                                         }

        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock,
                                             report_only=True)

        # two mount options added --> two info messages
        self.assertEqual(len(messages), 2)
        self.assertEqual(messages[0].type, common.MESSAGE_TYPE_INFO)
        self.assertEqual(messages[1].type, common.MESSAGE_TYPE_INFO)

        # newly added mount options should be mentioned in the messages
        # together with their mount points
        nodev_found = False
        noauto_found = False

        for message in messages:
            if "'nodev'" in message.text:
                self.assertIn("/tmp", message.text)
                nodev_found = True
            elif "'noauto'" in message.text:
                self.assertIn("/", message.text)
                noauto_found = True

        self.assertTrue(all([nodev_found, noauto_found]))

        # no changes should be made
        self.assertEqual(self.storage_mock.mountpoints["/tmp"].format.options,
                         "defaults")
        self.assertEqual(self.storage_mock.mountpoints["/"].format.options,
                         "defaults")

    def add_mount_option_prefix_test(self):
        for rule in ["part /tmp --mountoptions=nodev",
                     "part / --mountoptions=noauto"]:
            self.rule_data.new_rule(rule)

        tmp_part_mock = mock.Mock()
        tmp_part_mock.format.options = "defaults,nodevice"
        root_part_mock = mock.Mock()
        root_part_mock.format.options = "defaults"

        self.storage_mock.mountpoints = {"/tmp": tmp_part_mock,
                                         "/": root_part_mock,
                                         }

        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock)

        # two mount options added (even though it is a prefix of another one)
        #   --> two info messages
        self.assertEqual(len(messages), 2)
        self.assertEqual(messages[0].type, common.MESSAGE_TYPE_INFO)
        self.assertEqual(messages[1].type, common.MESSAGE_TYPE_INFO)

        # the option should be added even though it is a prefix of another one
        self.assertEqual(self.storage_mock.mountpoints["/tmp"].format.options,
                         "defaults,nodevice,nodev")

    def add_mount_options_nonexisting_part_test(self):
        for rule in ["part /tmp --mountoptions=nodev",
                     "part / --mountoptions=noauto"]:
            self.rule_data.new_rule(rule)

        tmp_part_mock = mock.Mock()
        tmp_part_mock.format.options = "defaults"
        root_part_mock = mock.Mock()
        root_part_mock.format.options = "defaults"

        self.storage_mock.mountpoints = {"/": root_part_mock,
                                         }

        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock)

        # one mount option added, one mount point missing (mount options
        # cannot be added) --> one info, one error
        self.assertEqual(len(messages), 2)
        self.assertTrue(any(message.type == common.MESSAGE_TYPE_INFO
                            for message in messages))
        self.assertTrue(any(message.type == common.MESSAGE_TYPE_FATAL
                            for message in messages))

        # the info message should report mount options added to the existing
        # mount point, the error message shoud contain the missing mount point
        # and not the mount option
        for message in messages:
            if message.type == common.MESSAGE_TYPE_INFO:
                self.assertIn("/", message.text)
                self.assertIn("'noauto'", message.text)
            elif message.type == common.MESSAGE_TYPE_FATAL:
                self.assertIn("/tmp", message.text)
                self.assertNotIn("'nodev'", message.text)

    def passwd_minlen_no_passwd_test(self):
        self.rule_data.new_rule("passwd --minlen=8")

        self.ksdata_mock.rootpw.password = ""
        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock)

        # minimal password length required --> one warning
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0].type, common.MESSAGE_TYPE_WARNING)

        # warning has to mention the length
        self.assertIn("8", messages[0].text)

    def passwd_minlen_short_passwd_test(self):
        self.rule_data.new_rule("passwd --minlen=8")

        self.ksdata_mock.rootpw.password = "aaaa"
        self.ksdata_mock.rootpw.isCrypted = False
        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock)

        # minimal password length greater than actual length --> one warning
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0].type, common.MESSAGE_TYPE_FATAL)

        # warning has to mention the length
        self.assertIn("8", messages[0].text)

        # warning should mention that something is wrong with the old password
        self.assertIn("is", messages[0].text)

        # doing changes --> password should not be cleared
        self.assertEqual(self.ksdata_mock.rootpw.password, "aaaa")

    def passwd_minlen_short_passwd_report_only_test(self):
        self.rule_data.new_rule("passwd --minlen=8")

        self.ksdata_mock.rootpw.password = "aaaa"
        self.ksdata_mock.rootpw.isCrypted = False

        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock,
                                             report_only=True)

        # minimal password length greater than actual length --> one warning
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0].type, common.MESSAGE_TYPE_FATAL)

        # warning has to mention the length
        self.assertIn("8", messages[0].text)

        # report only --> password shouldn't be cleared
        self.assertEqual(self.ksdata_mock.rootpw.password, "aaaa")

    def passwd_minlen_crypted_passwd_test(self):
        self.rule_data.new_rule("passwd --minlen=8")

        self.ksdata_mock.rootpw.password = "aaaa"
        self.ksdata_mock.rootpw.isCrypted = True

        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock)

        # minimal password length greater than actual length --> one warning
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0].type, common.MESSAGE_TYPE_WARNING)

        # warning has to mention that the password cannot be checked
        self.assertIn("cannot check", messages[0].text)

    def passwd_minlen_good_passwd_test(self):
        self.rule_data.new_rule("passwd --minlen=8")

        self.ksdata_mock.rootpw.password = "aaaaaaaaaaaaaaaaa"
        self.ksdata_mock.rootpw.isCrypted = False

        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock)

        # minimal password length less than actual length --> no warning
        self.assertEqual(messages, [])

    def passwd_minlen_report_only_not_ignored_test(self):
        self.rule_data.new_rule("passwd --minlen=8")
        self.ksdata_mock.rootpw.password = "aaaaaaaaaaaaaaaaa"
        self.ksdata_mock.rootpw.isCrypted = False

        # Mock pw_policy returned by anaconda.pwpolicy.get_policy()
        pw_policy_mock = mock.Mock()
        pw_policy_mock.minlen = 6
        pw_policy_mock.strict = False
        self.ksdata_mock.anaconda.pwpolicy.get_policy.return_value = \
            pw_policy_mock

        # call eval_rules with report_only=False
        # should set password minimal length to 8
        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock,
                                             report_only=False)
        # Password Policy changed --> no warnings
        self.assertEqual(messages, [])
        self.assertEqual(self.rule_data._passwd_rules._orig_minlen, 6)
        self.assertEqual(self.rule_data._passwd_rules._orig_strict, False)
        self.assertEqual(pw_policy_mock.minlen, 8)
        self.assertEqual(pw_policy_mock.strict, True)
        self.assertEqual(self.rule_data._passwd_rules._minlen, 8)

        # call of eval_rules with report_only=True
        # should not change anything
        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock,
                                             report_only=True)
        # Password Policy stayed the same --> no warnings
        self.assertEqual(messages, [])

        self.assertEqual(self.rule_data._passwd_rules._orig_minlen, 6)
        self.assertEqual(self.rule_data._passwd_rules._orig_strict, False)
        self.assertEqual(pw_policy_mock.minlen, 8)
        self.assertEqual(pw_policy_mock.strict, True)
        self.assertEqual(self.rule_data._passwd_rules._minlen, 8)

    def package_rules_test(self):
        self.rule_data.new_rule("package --add=firewalld --remove=telnet "
                                "--add=iptables --add=vim")

        self.ksdata_mock.packages.packageList = ["vim"]
        self.ksdata_mock.packages.excludedList = []

        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock)

        # one info message for each (really) added/removed package
        self.assertEqual(len(messages), 3)
        self.assertTrue(all(message.type == common.MESSAGE_TYPE_INFO
                            for message in messages))

        # all packages should appear in the messages
        must_see = ["firewalld", "telnet", "iptables"]
        for message in messages:
            if "'firewalld'" in message.text:
                must_see.remove("firewalld")
            elif "'telnet'" in message.text:
                must_see.remove("telnet")
            elif "'iptables'":
                must_see.remove("iptables")

        self.assertEqual(must_see, [])
        self.assertEqual(set(self.ksdata_mock.packages.packageList),
                         {"firewalld", "iptables", "vim"})
        self.assertEqual(set(self.ksdata_mock.packages.excludedList),
                         {"telnet"})

    def package_rules_report_only_test(self):
        self.rule_data.new_rule("package --add=firewalld --remove=telnet "
                                "--add=iptables")

        self.ksdata_mock.packages.packageList = []
        self.ksdata_mock.packages.excludedList = []

        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock,
                                             report_only=True)

        # one info message for each added/removed package
        self.assertEqual(len(messages), 3)
        self.assertTrue(all(message.type == common.MESSAGE_TYPE_INFO
                            for message in messages))

        # all packages should appear in the messages
        must_see = ["firewalld", "telnet", "iptables"]
        for message in messages:
            if "'firewalld'" in message.text:
                must_see.remove("firewalld")
            elif "'telnet'" in message.text:
                must_see.remove("telnet")
            elif "'iptables'":
                must_see.remove("iptables")

        self.assertEqual(must_see, [])

        # report_only --> no packages should be added or excluded
        self.assertEqual(self.ksdata_mock.packages.packageList, [])
        self.assertEqual(self.ksdata_mock.packages.excludedList, [])

    def bootloader_passwd_not_set_test(self):
        self.rule_data.new_rule("bootloader --passwd")

        self.storage_mock.bootloader.password = None

        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock)

        # bootloader password not set --> one warning
        self.assertEqual(len(messages), 1)
        self.assertTrue(messages[0].type == common.MESSAGE_TYPE_WARNING)

    def bootloader_passwd_set_test(self):
        self.rule_data.new_rule("bootloader --passwd")

        self.storage_mock.bootloader.password = "aaaaa"

        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock)

        # bootloader password set --> no warnings
        self.assertEqual(messages, [])

    def various_rules_test(self):
        for rule in ["part /tmp", "part /", "passwd --minlen=14",
                     "package --add=firewalld", ]:
            self.rule_data.new_rule(rule)

        self.storage_mock.mountpoints = dict()
        self.ksdata_mock.packages.packageList = []
        self.ksdata_mock.packages.excludedList = []

        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock)

        # four rules, all fail --> four messages
        self.assertEqual(len(messages), 4)


class RevertingTest(unittest.TestCase):
    """Test for reverting changes done by the rule evaluation."""

    def setUp(self):
        self.rule_data = rule_handling.RuleData()
        self.ksdata_mock = mock.Mock()
        self.storage_mock = mock.Mock()

    def revert_mount_options_nonexistent_test(self):
        self.rule_data.new_rule("part /tmp --mountoptions=nodev")
        self.storage_mock.mountpoints = dict()

        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock)

        # mount point doesn't exist -> one message, nothing done
        self.assertEqual(len(messages), 1)
        self.assertEqual(self.storage_mock.mountpoints, dict())

        # mount point doesn't exist -> shouldn't do anything
        self.rule_data.revert_changes(self.ksdata_mock, self.storage_mock)
        self.assertEqual(self.storage_mock.mountpoints, dict())

    def revert_mount_options_test(self):
        self.rule_data.new_rule("part /tmp --mountoptions=nodev")
        self.storage_mock.mountpoints = dict()
        self.storage_mock.mountpoints["/tmp"] = mock.Mock()
        self.storage_mock.mountpoints["/tmp"].format.options = "defaults"

        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock)

        # mount option added --> one message
        self.assertEqual(len(messages), 1)

        # "nodev" option should be added
        self.assertEqual(self.storage_mock.mountpoints["/tmp"].format.options,
                         "defaults,nodev")

        self.rule_data.revert_changes(self.ksdata_mock, self.storage_mock)

        # should be reverted to the original value
        self.assertEqual(self.storage_mock.mountpoints["/tmp"].format.options,
                         "defaults")

        # another cycle of the same #
        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock)

        # mount option added --> one message
        self.assertEqual(len(messages), 1)

        # "nodev" option should be added
        self.assertEqual(self.storage_mock.mountpoints["/tmp"].format.options,
                         "defaults,nodev")

        self.rule_data.revert_changes(self.ksdata_mock, self.storage_mock)

        # should be reverted to the original value
        self.assertEqual(self.storage_mock.mountpoints["/tmp"].format.options,
                         "defaults")

    def revert_password_policy_changes_test(self):
        # FIXME: Add password policy changes to this test. It only checks
        # password length right now outside of policy changes.
        self.rule_data.new_rule("passwd --minlen=8")

        self.ksdata_mock.rootpw.password = "aaaa"
        self.ksdata_mock.rootpw.isCrypted = False
        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock)

        # password error --> one message
        self.assertEqual(len(messages), 1)
        self.assertEqual(self.ksdata_mock.rootpw.password, "aaaa")
        self.assertTrue(self.ksdata_mock.rootpw.seen)

        self.rule_data.revert_changes(self.ksdata_mock, self.storage_mock)

        # with long enough password this time #
        self.ksdata_mock.rootpw.password = "aaaaaaaaaaaaa"

        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock)

        # long enough password
        # entered --> no message
        self.assertEqual(messages, [])

    def revert_package_rules_test(self):
        self.rule_data.new_rule("package --add=firewalld --remove=telnet "
                                "--add=iptables --add=vim")

        self.ksdata_mock.packages.packageList = ["vim"]
        self.ksdata_mock.packages.excludedList = []

        # run twice --> nothing should be different in the second run
        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock)
        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock)

        # one info message for each added/removed package
        self.assertEqual(len(messages), 3)

        self.rule_data.revert_changes(self.ksdata_mock, self.storage_mock)

        # (only) added and excluded packages should have been removed from the
        # list
        self.assertEqual(self.ksdata_mock.packages.packageList, ["vim"])
        self.assertEqual(self.ksdata_mock.packages.excludedList, [])

        # now do the same again #
        messages = self.rule_data.eval_rules(self.ksdata_mock,
                                             self.storage_mock)

        # one info message for each added/removed package
        self.assertEqual(len(messages), 3)

        self.rule_data.revert_changes(self.ksdata_mock, self.storage_mock)

        # (only) added and excluded packages should have been removed from the
        # list
        self.assertEqual(self.ksdata_mock.packages.packageList, ["vim"])
        self.assertEqual(self.ksdata_mock.packages.excludedList, [])



#
# Copyright (C) 2017  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Watson Sato <wsato@redhat.com>
#

"""Module with unit tests for the content_handling.py module"""

import unittest
import mock

from org_fedora_oscap import common
from org_fedora_oscap import content_handling as ch


class DataStreamHandlerTest(unittest.TestCase):
    """Test functionality of the DataStreanHandler'."""

    def setUp(self):
        self.ds_filepath = "../testing_files/testing_ds.xml"
        self.ds_ids = "scap_org.open-scap_datastream_tst"
        self.chk_first_id = "scap_org.open-scap_cref_first-xccdf.xml"
        self.chk_second_id = "scap_org.open-scap_cref_second-xccdf.xml"
        self.profile1_id = "xccdf_com.example_profile_my_profile"
        self.profile2_id = "xccdf_com.example_profile_my_profile2"
        self.profile3_id = "xccdf_com.example_profile_my_profile3"

    def init_invalid_file_path_test(self):
        with self.assertRaises(ch.DataStreamHandlingError) as e:
            ch.DataStreamHandler("testing_ds.xml")
        self.assertIn("Invalid file path", e.exception.message)

    def init_not_scap_content_test(self):
        with self.assertRaises(ch.DataStreamHandlingError) as e:
            ch.DataStreamHandler("../testing_files/testing_ks.cfg")
        self.assertIn("not a valid SCAP content file", e.exception.message)

    def init_xccdf_content_test(self):
        with self.assertRaises(ch.DataStreamHandlingError) as e:
            ch.DataStreamHandler("../testing_files/xccdf.xml")
        self.assertIn("not a data stream collection", e.exception.message)

    def get_data_streams_test(self):
        self.ds_handler = ch.DataStreamHandler(self.ds_filepath)
        ds_ids = self.ds_handler.get_data_streams()
        self.assertIn(self.ds_ids, ds_ids)

    def get_data_streams_checklists_test(self):
        expected_ids = {self.ds_ids: [self.chk_first_id, self.chk_second_id]}

        self.ds_handler = ch.DataStreamHandler(self.ds_filepath)
        ds_ids = self.ds_handler.get_data_streams_checklists()
        self.assertDictEqual(expected_ids, ds_ids)

    def get_checklists_test(self):
        expected_checklists = [self.chk_first_id, self.chk_second_id]

        self.ds_handler = ch.DataStreamHandler(self.ds_filepath)
        chk_ids = self.ds_handler.get_checklists(self.ds_ids)
        self.assertListEqual(expected_checklists, chk_ids)

    def get_checklists_invalid_test(self):
        self.ds_handler = ch.DataStreamHandler(self.ds_filepath)
        with self.assertRaises(ch.DataStreamHandlingError) as e:
            chk_ids = self.ds_handler.get_checklists("invalid.id")
        self.assertIn("Invalid data stream id given", e.exception.message)

    def get_profiles_test(self):
        self.ds_handler = ch.DataStreamHandler(self.ds_filepath)
        profile_ids = self.ds_handler.get_profiles(self.ds_ids,
                                                   self.chk_first_id)

        # When Benchmark doesn't contain Rules selected by default
        # the default Profile should not be present
        self.assertEqual(2, len(profile_ids))
        self.assertEqual(self.profile1_id, profile_ids[0].id)
        self.assertEqual(self.profile2_id, profile_ids[1].id)

    def get_profiles_with_default(self):
        self.ds_handler = ch.DataStreamHandler(self.ds_filepath)
        profile_ids = self.ds_handler.get_profiles(self.ds_ids,
                                                   self.chk_second_id)

        # When Benchmark contains Rules selected by default
        # the default Profile should be present
        self.assertEqual(2, len(profile_ids))
        self.assertEqual("default", profile_ids[0].id)
        self.assertEqual(self.profile3_id, profile_ids[1].id)



#
# Copyright (C) 2013  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Vratislav Podzimek <vpodzime@redhat.com>
#

"""
Module with various classes for handling pre-installation rules.

"""

import optparse
import shlex
import gettext
import logging
from pyanaconda.pwpolicy import F22_PwPolicyData
from org_fedora_oscap import common
from org_fedora_oscap.common import OSCAPaddonError, RuleMessage

# everything else should be private
__all__ = ["RuleData"]

_ = lambda x: gettext.ldgettext("oscap-anaconda-addon", x)

log = logging.getLogger("anaconda")


# TODO: use set instead of list for mount options?
def parse_csv(option, opt_str, value, parser):
    for item in value.split(","):
        if item:
            parser.values.ensure_value(option.dest, []).append(item)


class ModifiedOptionParserException(Exception):
    """Exception to be raised by ModifiedOptionParser."""
    pass


class ModifiedOptionParser(optparse.OptionParser):
    """Overrides error behavior of OptionParser."""
    def error(self, msg):
        raise ModifiedOptionParserException(msg)

    def exit(self, status=0, msg=None):
        raise ModifiedOptionParserException(msg)


PART_RULE_PARSER = ModifiedOptionParser()
PART_RULE_PARSER.add_option("--mountoptions", dest="mount_options",
                            action="callback", callback=parse_csv, nargs=1,
                            type="string")

PASSWD_RULE_PARSER = ModifiedOptionParser()
PASSWD_RULE_PARSER.add_option("--minlen", dest="minlen", action="store",
                              default=0, type="int")

PACKAGE_RULE_PARSER = ModifiedOptionParser()
PACKAGE_RULE_PARSER.add_option("--add", dest="add_pkgs", action="append",
                               type="string")
PACKAGE_RULE_PARSER.add_option("--remove", dest="remove_pkgs", action="append",
                               type="string")

BOOTLOADER_RULE_PARSER = ModifiedOptionParser()
BOOTLOADER_RULE_PARSER.add_option("--passwd", dest="passwd", action="store_true",
                                  default=False)

KDUMP_RULE_PARSER = ModifiedOptionParser()
KDUMP_RULE_PARSER.add_option("--enable", action="store_true",
                             dest="kdenabled", default=None)
KDUMP_RULE_PARSER.add_option("--disable", action="store_false",
                             dest="kdenabled", default=None)

FIREWALL_RULE_PARSER = ModifiedOptionParser()
FIREWALL_RULE_PARSER.add_option("--enable", action="store_true",
                                dest="fwenabled", default=None)
FIREWALL_RULE_PARSER.add_option("--disable", action="store_false",
                                dest="fwenabled", default=None)
FIREWALL_RULE_PARSER.add_option("--service", dest="add_svcs", action="append",
                                type="string")
FIREWALL_RULE_PARSER.add_option("--port", dest="add_port", action="append",
                                type="string")
FIREWALL_RULE_PARSER.add_option("--trust", dest="add_trust", action="append",
                                type="string")
FIREWALL_RULE_PARSER.add_option("--remove-service", dest="remove_svcs",
                                action="append", type="string")


class RuleHandler(object):
    """Base class for the rule handlers."""

    def eval_rules(self, ksdata, storage, report_only=False):
        """
        Method that should check the current state (as defined by the ksdata and
        storage parameters) against the rules the instance of RuleHandler
        holds. Depending on the value of report_only it should fix the state
        with changes that can be done automatically or not and return the list
        of warnings and errors for fixes that need to be done manually together
        with info messages about the automatic changes. One should make sure
        this method is called with report_only set to False at least once so
        that the automatic fixes are done.

        :param ksdata: data representing the values set by user
        :type ksdata: pykickstart.base.BaseHandler
        :param storage: object storing storage-related information
                        (disks, partitioning, bootloader, etc.)
        :type storage: blivet.Blivet
        :param report_only: whether to do fixing or just report information
        :type report_only: bool
        :return: errors and warnings for fixes that need to be done manually and
                 info messages about the automatic changes
        :rtype: list of common.RuleMessage objects

        """

        return []

    def revert_changes(self, ksdata, storage):
        """
        Method that should revert all changes done by the previous calls of the
        eval_rules method with the report_only set to False.

        :see: eval_rules

        """

        # inheriting classes are supposed to override this
        pass


class UknownRuleError(OSCAPaddonError):
    """Exception class for cases when an uknown rule is to be processed."""

    pass


class RuleData(RuleHandler):
    """Class holding data parsed from the applied rules."""

    def __init__(self):
        """Constructor initializing attributes."""

        self._part_rules = PartRules()
        self._passwd_rules = PasswdRules()
        self._package_rules = PackageRules()
        self._bootloader_rules = BootloaderRules()
        self._kdump_rules = KdumpRules()
        self._firewall_rules = FirewallRules()

        self._rule_handlers = (self._part_rules, self._passwd_rules,
                               self._package_rules, self._bootloader_rules,
                               self._kdump_rules, self._firewall_rules,
                               )

    def __str__(self):
        """Standard method useful for debugging and testing."""

        ret = ""

        part_strs = str(self._part_rules)
        if part_strs:
            ret += part_strs

        passwd_str = str(self._passwd_rules)
        if passwd_str:
            ret += "\n" + passwd_str

        packages_str = str(self._package_rules)
        if packages_str:
            ret += "\n" + packages_str

        firewall_str = str(self._firewall_rules)
        if firewall_str:
            ret += "\n" + firewall_str

        return ret

    def new_rule(self, rule):
        """
        Method that handles a single rule line (e.g. "part /tmp").

        :param rule: a single rule line
        :type rule: str

        """

        actions = {"part": self._new_part_rule,
                   "passwd": self._new_passwd_rule,
                   "package": self._new_package_rule,
                   "bootloader": self._new_bootloader_rule,
                   "kdump": self._new_kdump_rule,
                   "firewall": self._new_firewall_rule,
                   }

        rule = rule.strip()
        if not rule:
            return

        first_word = rule.split(None, 1)[0]
        try:
            actions[first_word](rule)
        except (ModifiedOptionParserException, KeyError) as e:
            log.warning("Unknown OSCAP Addon rule '{}': {}".format(rule, e))

    def eval_rules(self, ksdata, storage, report_only=False):
        """:see: RuleHandler.eval_rules"""

        messages = []

        # evaluate all subgroups of rules
        for rule_handler in self._rule_handlers:
            messages += rule_handler.eval_rules(ksdata, storage, report_only)

        return messages

    def revert_changes(self, ksdata, storage):
        """:see: RuleHandler.revert_changes"""

        # revert changes in all subgroups of rules
        for rule_handler in self._rule_handlers:
            rule_handler.revert_changes(ksdata, storage)

    def _new_part_rule(self, rule):
        args = shlex.split(rule)
        (opts, args) = PART_RULE_PARSER.parse_args(args)

        # args contain both "part" and mount point (e.g. "/tmp")
        mount_point = args[1]

        self._part_rules.ensure_mount_point(mount_point)

        if opts.mount_options:
            part_data = self._part_rules[mount_point]
            part_data.add_mount_options(opts.mount_options)

    def _new_passwd_rule(self, rule):
        args = shlex.split(rule)
        (opts, args) = PASSWD_RULE_PARSER.parse_args(args)

        self._passwd_rules.update_minlen(opts.minlen)

    def _new_package_rule(self, rule):
        args = shlex.split(rule)
        (opts, args) = PACKAGE_RULE_PARSER.parse_args(args)

        self._package_rules.add_packages(opts.add_pkgs)
        self._package_rules.remove_packages(opts.remove_pkgs)

    def _new_bootloader_rule(self, rule):
        args = shlex.split(rule)
        (opts, args) = BOOTLOADER_RULE_PARSER.parse_args(args)

        if opts.passwd:
            self._bootloader_rules.require_password()

    def _new_kdump_rule(self, rule):
        args = shlex.split(rule)
        (opts, args) = KDUMP_RULE_PARSER.parse_args(args)

        self._kdump_rules.kdump_enabled(opts.kdenabled)

    def _new_firewall_rule(self, rule):
        args = shlex.split(rule)
        (opts, args) = FIREWALL_RULE_PARSER.parse_args(args)

        self._firewall_rules.add_services(opts.add_svcs)
        self._firewall_rules.remove_services(opts.remove_svcs)
        self._firewall_rules.add_trusts(opts.add_trust)
        self._firewall_rules.add_ports(opts.add_port)
        self._firewall_rules.firewall_enabled(opts.fwenabled)

    @property
    def passwd_rules(self):
        # needed for fixups in GUI
        return self._passwd_rules


class PartRules(RuleHandler):
    """Simple class holding data from the rules affecting partitioning."""

    def __init__(self):
        """Constructor initializing attributes."""

        self._rules = dict()

    def __str__(self):
        """Standard method useful for debugging and testing."""

        return "\n".join(str(rule) for rule in self._rules.itervalues())

    def __getitem__(self, key):
        """Method to support dictionary-like syntax."""

        return self._rules[key]

    def __setitem__(self, key, value):
        """Method to support dictionary-like syntax."""

        self._rules[key] = value

    def __delitem__(self, key):
        """One of the methods needed to implement a container."""

        self._rules.__delitem__(key)

    def __len__(self):
        """One of the methods needed to implement a container."""

        return self._rules.__len__()

    def __contains__(self, key):
        """Method needed for the 'in' operator to work."""

        return key in self._rules

    def ensure_mount_point(self, mount_point):
        if mount_point not in self._rules:
            self._rules[mount_point] = PartRule(mount_point)

    def eval_rules(self, ksdata, storage, report_only=False):
        """:see: RuleHandler.eval_rules"""

        messages = []
        for part_rule in self._rules.itervalues():
            messages += part_rule.eval_rules(ksdata, storage, report_only)

        return messages

    def revert_changes(self, ksdata, storage):
        """:see: RuleHandler.revert_changes"""

        for part_rule in self._rules.itervalues():
            part_rule.revert_changes(ksdata, storage)


class PartRule(RuleHandler):
    """Simple class holding rule data for a single partition/mount point."""

    def __init__(self, mount_point):
        """
        Constructor initializing attributes.

        :param mount_point: the mount point the object holds data for
        :type mount_point: str

        """

        self._mount_point = mount_point
        self._mount_options = []
        self._added_mount_options = []

    def __str__(self):
        """Standard method useful for debugging and testing."""

        ret = "part %s" % self._mount_point
        if self._mount_options:
            ret += " --mountoptions=%s" % ",".join(self._mount_options)

        return ret

    def add_mount_options(self, mount_options):
        """
        Add  new mount options (do not add duplicates).

        :param mount_options: list of mount options to be added
        :type mount_options: list of strings

        """

        self._mount_options.extend(opt for opt in mount_options
                                   if opt not in self._mount_options)

    def eval_rules(self, ksdata, storage, report_only=False):
        """:see: RuleHandler.eval_rules"""

        messages = []
        if self._mount_point not in storage.mountpoints:
            msg = _("{0} must be on a separate partition or logical "
                    "volume and has to be created in the "
                    "partitioning layout before installation can occur "
                    "with a security profile").format(self._mount_point)
            messages.append(RuleMessage(self.__class__,
                                        common.MESSAGE_TYPE_FATAL, msg))

            # mount point doesn't exist, nothing more can be found here
            return messages

        # template for the message
        msg_tmpl = _("mount option '%(mount_option)s' added for "
                     "the mount point %(mount_point)s")

        # add message for every option already added
        for opt in self._added_mount_options:
            msg = msg_tmpl % {"mount_option": opt,
                              "mount_point": self._mount_point}
            messages.append(RuleMessage(self.__class__,
                                        common.MESSAGE_TYPE_INFO, msg))

        # mount point to be created during installation
        target_mount_point = storage.mountpoints[self._mount_point]

        # generator for the new options that should be added
        new_opts = (opt for opt in self._mount_options
                    if opt not in target_mount_point.format.options.split(","))

        # add message for every mount option added
        for opt in new_opts:
            msg = msg_tmpl % {"mount_option": opt,
                              "mount_point": self._mount_point}

            # add message for the mount option in any case
            messages.append(RuleMessage(self.__class__,
                                        common.MESSAGE_TYPE_INFO, msg))

            # add new options to the target mount point if not reporting only
            if not report_only:
                target_mount_point.format.options += ",%s" % opt
                self._added_mount_options.append(opt)

        return messages

    def revert_changes(self, ksdata, storage):
        """
        Removes the mount options added to the mount point by this PartRule
        instance.

        :see: RuleHandler.revert_changes

        """

        if self._mount_point not in storage.mountpoints:
            # mount point doesn't exist, nothing can be reverted
            return

        # mount point to be created during installation
        target_mount_point = storage.mountpoints[self._mount_point]

        # mount options to be defined for the created mount point
        tgt_mount_options = target_mount_point.format.options

        # generator of the options that should remain
        result_opts = (opt for opt in tgt_mount_options.split(",")
                       if opt not in self._added_mount_options)

        # set the new list of options
        target_mount_point.format.options = ",".join(result_opts)

        # reset the remembered added mount options
        self._added_mount_options = []


class PasswdRules(RuleHandler):
    """Simple class holding data from the rules affecting passwords."""

    def __init__(self):
        """Constructor initializing attributes."""

        self._minlen = 0
        self._created_policy = False
        self._orig_minlen = None
        self._orig_strict = None

    def __str__(self):
        """Standard method useful for debugging and testing."""

        if self._minlen > 0:
            return "passwd --minlen=%d" % self._minlen
        else:
            return ""

    def update_minlen(self, minlen):
        """Update password minimal length requirements."""

        if minlen > self._minlen:
            self._minlen = minlen

    def eval_rules(self, ksdata, storage, report_only=False):
        """:see: RuleHandler.eval_rules"""

        if self._minlen == 0:
            # no password restrictions, nothing to be done here
            return []

        ret = []
        if not ksdata.rootpw.password:
            # root password was not set

            msg = _("make sure to create password with minimal length of %d "
                    "characters") % self._minlen
            ret = [RuleMessage(self.__class__,
                               common.MESSAGE_TYPE_WARNING, msg)]
        else:
            # root password set
            if ksdata.rootpw.isCrypted:
                msg = _("cannot check root password length (password is crypted)")
                log.warning("cannot check root password length (password is crypted)")
                return [RuleMessage(self.__class__,
                                    common.MESSAGE_TYPE_WARNING, msg)]
            elif len(ksdata.rootpw.password) < self._minlen:
                # too short
                msg = _("root password is too short, a longer one with at "
                        "least %d characters is required") % self._minlen
                ret = [RuleMessage(self.__class__,
                                   common.MESSAGE_TYPE_FATAL, msg)]
            else:
                ret = []

        if report_only:
            return ret

        # set the policy in any case (so that a weaker password is not entered)
        pw_policy = ksdata.anaconda.pwpolicy.get_policy("root")
        if pw_policy is None:
            pw_policy = F22_PwPolicyData()
            log.info("OSCAP addon: setting password policy %s" % pw_policy)
            ksdata.anaconda.pwpolicy.policyList.append(pw_policy)
            log.info("OSCAP addon: password policy list: %s" % ksdata.anaconda.pwpolicy.policyList)
            self._created_policy = True

        self._orig_minlen = pw_policy.minlen
        self._orig_strict = pw_policy.strict
        pw_policy.minlen = self._minlen
        pw_policy.strict = True

        return ret

    def revert_changes(self, ksdata, storage):
        """:see: RuleHander.revert_changes"""

        pw_policy = ksdata.anaconda.pwpolicy.get_policy("root")
        if self._created_policy:
            log.info("OSCAP addon: removing password policy: %s" % pw_policy)
            ksdata.anaconda.pwpolicy.policyList.remove(pw_policy)
            log.info("OSCAP addon: password policy list: %s" % ksdata.anaconda.pwpolicy.policyList)
            self._created_policy = False
        else:
            if self._orig_minlen is not None:
                pw_policy.minlen = self._orig_minlen
                self._orig_minlen = None
            if self._orig_strict is not None:
                pw_policy.strict = self._orig_strict
                self._orig_strict = None


class PackageRules(RuleHandler):
    """Simple class holding data from the rules affecting installed packages."""

    def __init__(self):
        """Constructor setting the initial value of attributes."""

        self._add_pkgs = set()
        self._remove_pkgs = set()

        self._added_pkgs = set()
        self._removed_pkgs = set()

    def add_packages(self, packages):
        """
        New packages that should be added.

        :param packages: packages to be added
        :type packages: iterable

        """

        if packages:
            self._add_pkgs.update(packages)

    def remove_packages(self, packages):
        """
        New packages that should be removed.

        :param packages: packages to be removed
        :type packages: iterable

        """

        if packages:
            self._remove_pkgs.update(packages)

    def __str__(self):
        """Standard method useful for debugging and testing."""

        ret = "packages"
        adds = " ".join("--add=%s" % package for package in self._add_pkgs)
        if adds:
            ret += " " + adds

        rems = " ".join("--remove=%s" % package
                        for package in self._remove_pkgs)
        if rems:
            ret += " " + rems

        return ret

    def eval_rules(self, ksdata, storage, report_only=False):
        """:see: RuleHandler.eval_rules"""

        messages = []

        # add messages for the already added packages
        for pkg in self._added_pkgs:
            msg = _("package '%s' has been added to the list of to be installed "
                    "packages" % pkg)
            messages.append(RuleMessage(self.__class__,
                                        common.MESSAGE_TYPE_INFO, msg))

        # packages, that should be added
        packages_to_add = (pkg for pkg in self._add_pkgs
                           if pkg not in ksdata.packages.packageList)

        for pkg in packages_to_add:
            # add the package unless already added
            if not report_only:
                self._added_pkgs.add(pkg)
                ksdata.packages.packageList.append(pkg)

            msg = _("package '%s' has been added to the list of to be installed "
                    "packages" % pkg)
            messages.append(RuleMessage(self.__class__,
                                        common.MESSAGE_TYPE_INFO, msg))

        # now do the same for the packages that should be excluded

        # add messages for the already excluded packages
        for pkg in self._removed_pkgs:
            msg = _("package '%s' has been added to the list of excluded "
                    "packages" % pkg)
            messages.append(RuleMessage(self.__class__,
                                        common.MESSAGE_TYPE_INFO, msg))

        # packages, that should be added
        packages_to_remove = (pkg for pkg in self._remove_pkgs
                              if pkg not in ksdata.packages.excludedList)

        for pkg in packages_to_remove:
            # exclude the package unless already excluded
            if not report_only:
                self._removed_pkgs.add(pkg)
                ksdata.packages.excludedList.append(pkg)

            msg = _("package '%s' has been added to the list of excluded "
                    "packages" % pkg)
            messages.append(RuleMessage(self.__class__,
                                        common.MESSAGE_TYPE_INFO, msg))

        return messages

    def revert_changes(self, ksdata, storage):
        """:see: RuleHander.revert_changes"""

        # remove all packages this handler added
        for pkg in self._added_pkgs:
            if pkg in ksdata.packages.packageList:
                ksdata.packages.packageList.remove(pkg)

        # remove all packages this handler excluded
        for pkg in self._removed_pkgs:
            if pkg in ksdata.packages.excludedList:
                ksdata.packages.excludedList.remove(pkg)

        self._added_pkgs = set()
        self._removed_pkgs = set()


class BootloaderRules(RuleHandler):
    """Simple class holding data from the rules affecting bootloader."""

    def __init__(self):
        """Constructor setting the initial value of attributes."""

        self._require_password = False

    def require_password(self):
        """Requests the bootloader password should be required."""

        self._require_password = True

    def __str__(self):
        """Standard method useful for debugging and testing."""

        ret = "bootloader"

        if self._require_password:
            ret += " --passwd"

        return ret

    def eval_rules(self, ksdata, storage, report_only=False):
        """:see: RuleHandler.eval_rules"""

        if self._require_password and not storage.bootloader.password:
            # Anaconda doesn't provide a way to set bootloader password, so
            # users cannot do much about that --> we shouldn't stop the
            # installation, should we?
            return [RuleMessage(self.__class__, common.MESSAGE_TYPE_WARNING,
                                "boot loader password not set up")]
        else:
            return []

    # nothing to be reverted for now


class KdumpRules(RuleHandler):
    """Simple class holding data from the rules affecting the kdump addon."""

    def __init__(self):
        """Constructor setting the initial value of attributes."""

        self._kdump_enabled = None
        self._kdump_default_enabled = None

    def kdump_enabled(self, kdenabled):
        """Enable or Disable Kdump"""

        if kdenabled is not None:
            self._kdump_enabled = kdenabled

    def __str__(self):
        """Standard method useful for debugging and testing."""

        ret = "kdump"

        if self._kdump_enabled is True:
            ret += " --enable"

        if self._kdump_enabled is False:
            ret += " --disable"

        return ret

    def eval_rules(self, ksdata, storage, report_only=False):
        """:see: RuleHandler.eval_rules"""

        messages = []

        if self._kdump_enabled is None:
            return []
        elif self._kdump_enabled is False:
            msg = _("Kdump will be disabled on startup")
        elif self._kdump_enabled is True:
            msg = _("Kdump will be enabled on startup")

        messages.append(RuleMessage(self.__class__,
                                    common.MESSAGE_TYPE_INFO, msg))

        if not report_only:
            try:
                if self._kdump_default_enabled is None:
                    # Kdump addon default startup setting
                    self._kdump_default_enabled = ksdata.addons.com_redhat_kdump.enabled
                ksdata.addons.com_redhat_kdump.enabled = self._kdump_enabled
            except AttributeError:
                log.warning("com_redhat_kdump is not installed. "
                            "Skipping kdump configuration")

        return messages

    def revert_changes(self, ksdata, storage):
        """:see: RuleHander.revert_changes"""

        try:
            if self._kdump_enabled is not None:
                ksdata.addons.com_redhat_kdump.enabled = self._kdump_default_enabled
        except AttributeError:
            log.warning("com_redhat_kdump is not installed. "
                        "Skipping reverting kdump configuration")

        self._kdump_enabled = None
        self._kdump_default_enabled = None


class FirewallRules(RuleHandler):
    """Simple class holding data from the rules affecting firewall configurations."""

    def __init__(self):
        """Constructor setting the initial value of attributes."""

        self._add_svcs = set()
        self._remove_svcs = set()
        self._add_trusts = set()
        self._add_ports = set()

        self._added_svcs = set()
        self._added_ports = set()
        self._added_trusts = set()
        self._removed_svcs = set()

        self._firewall_enabled = None
        self._firewall_default_enabled = None

    def add_services(self, services):
        """
        Services that should be allowed through firewall.

        :param services: services to be added
        :type services: iterable

        """

        if services:
            self._add_svcs.update(services)

    def add_ports(self, ports):
        """
        Ports that should be allowed through firewall.

        :param ports: ports to be added
        :type ports: iterable

        """

        if ports:
            self._add_ports.update(ports)

    def add_trusts(self, trusts):
        """
        trusts that should be allowed through firewall.

        :param trusts: trusts to be added
        :type trusts: iterable

        """

        if trusts:
            self._add_trusts.update(trusts)

    def remove_services(self, services):
        """
        New services that should not be allowed through firewall.

        :param services: services to be removed
        :type services: iterable

        """

        if services:
            self._remove_svcs.update(services)

    def firewall_enabled(self, fwenabled):
        """Enable or disable firewall"""

        if fwenabled is not None:
            self._firewall_enabled = fwenabled

    def __str__(self):
        """Standard method useful for debugging and testing."""

        ret = "firewall"

        if self._firewall_enabled is True:
            ret += " --enable"

        if self._firewall_enabled is False:
            ret += " --disable"

        adds = " ".join("--service=%s" % service
                        for service in self._add_svcs)
        if adds:
            ret += " " + adds

        rems = " ".join("--remove-service=%s" % service
                        for service in self._remove_svcs)
        if rems:
            ret += " " + rems

        ports = " ".join("--port=%s" % port
                         for port in self._add_ports)
        if ports:
            ret += " " + ports

        trusts = " ".join("--trust=%s" % trust
                          for trust in self._add_trusts)
        if trusts:
            ret += " " + trusts

        return ret

    def eval_rules(self, ksdata, storage, report_only=False):
        """:see: RuleHandler.eval_rules"""

        messages = []

        if self._firewall_default_enabled is None:
            # firewall default startup setting
            self._firewall_default_enabled = ksdata.firewall.enabled

        if self._firewall_enabled is False:
            msg = _("Firewall will be disabled on startup")
            messages.append(RuleMessage(self.__class__,
                                        common.MESSAGE_TYPE_INFO, msg))
            if not report_only:
                ksdata.firewall.enabled = self._firewall_enabled

        elif self._firewall_enabled is True:
            msg = _("Firewall will be enabled on startup")
            messages.append(RuleMessage(self.__class__,
                                        common.MESSAGE_TYPE_INFO, msg))
            if not report_only:
                ksdata.firewall.enabled = self._firewall_enabled

        # add messages for the already added services
        for svc in self._added_svcs:
            msg = _("service '%s' has been added to the list of services to be "
                    "added to the firewall" % svc)
            messages.append(RuleMessage(self.__class__,
                                        common.MESSAGE_TYPE_INFO, msg))

        # add messages for the already added ports
        for port in self._added_ports:
            msg = _("port '%s' has been added to the list of ports to be "
                    "added to the firewall" % port)
            messages.append(RuleMessage(self.__class__,
                                        common.MESSAGE_TYPE_INFO, msg))

        # add messages for the already added trusts
        for trust in self._added_trusts:
            msg = _("trust '%s' has been added to the list of trusts to be "
                    "added to the firewall" % trust)
            messages.append(RuleMessage(self.__class__,
                                        common.MESSAGE_TYPE_INFO, msg))

        # services, that should be added
        services_to_add = (svc for svc in self._add_svcs
                           if svc not in ksdata.firewall.services)

        # ports, that should be added
        ports_to_add = (ports for ports in self._add_ports
                        if ports not in ksdata.firewall.ports)

        # trusts, that should be added
        trusts_to_add = (trust for trust in self._add_trusts
                         if trust not in ksdata.firewall.trusts)

        for svc in services_to_add:
            # add the service unless already added
            if not report_only:
                self._added_svcs.add(svc)
                ksdata.firewall.services.append(svc)

            msg = _("service '%s' has been added to the list of services to be "
                    "added to the firewall" % svc)
            messages.append(RuleMessage(self.__class__,
                                        common.MESSAGE_TYPE_INFO, msg))

        for port in ports_to_add:
            # add the port unless already added
            if not report_only:
                self._added_ports.add(port)
                ksdata.firewall.ports.append(port)

            msg = _("port '%s' has been added to the list of ports to be "
                    "added to the firewall" % port)
            messages.append(RuleMessage(self.__class__,
                                        common.MESSAGE_TYPE_INFO, msg))

        for trust in trusts_to_add:
            # add the trust unless already added
            if not report_only:
                self._added_trusts.add(trust)
                ksdata.firewall.trusts.append(trust)

            msg = _("trust '%s' has been added to the list of trusts to be "
                    "added to the firewall" % trust)
            messages.append(RuleMessage(self.__class__,
                                        common.MESSAGE_TYPE_INFO, msg))

        # now do the same for the services that should be excluded

        # add messages for the already excluded services
        for svc in self._removed_svcs:
            msg = _("service '%s' has been added to the list of services to be "
                    "removed from the firewall" % svc)
            messages.append(RuleMessage(self.__class__,
                                        common.MESSAGE_TYPE_INFO, msg))

        # services, that should be added
        services_to_remove = (svc for svc in self._remove_svcs
                              if svc not in ksdata.firewall.remove_services)

        for svc in services_to_remove:
            # exclude the service unless already excluded
            if not report_only:
                self._removed_svcs.add(svc)
                ksdata.firewall.remove_services.append(svc)

            msg = _("service '%s' has been added to the list of services to be "
                    "removed from the firewall" % svc)
            messages.append(RuleMessage(self.__class__,
                                        common.MESSAGE_TYPE_INFO, msg))

        return messages

    def revert_changes(self, ksdata, storage):
        """:see: RuleHander.revert_changes"""

        if self._firewall_enabled is not None:
            ksdata.firewall.enabled = self._firewall_default_enabled

        # remove all services this handler added
        for svc in self._added_svcs:
            if svc in ksdata.firewall.services:
                ksdata.firewall.services.remove(svc)

        # remove all ports this handler added
        for port in self._added_ports:
            if port in ksdata.firewall.ports:
                ksdata.firewall.ports.remove(port)

        # remove all trusts this handler added
        for trust in self._added_trusts:
            if trust in ksdata.firewall.trusts:
                ksdata.firewall.trusts.remove(trust)

        # remove all services this handler excluded
        for svc in self._removed_svcs:
            if svc in ksdata.firewall.remove_services:
                ksdata.firewall.remove_services.remove(svc)

        self._added_svcs = set()
        self._added_ports = set()
        self._added_trusts = set()
        self._removed_svcs = set()
        self._firewall_enabled = None
        self._firewall_default_enabled = None



#
# Copyright (C) 2013  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Vratislav Podzimek <vpodzime@redhat.com>
#

"""Module with unit tests for the common.py module"""

import unittest
import os
import mock
from org_fedora_oscap import common


class OSCAPtoolRunningTest(unittest.TestCase):
    def setUp(self):
        self.mock_subprocess = mock.Mock()
        self.mock_subprocess.Popen = mock.Mock()
        self.mock_popen = mock.Mock()
        self.mock_communicate = mock.Mock()

        self.mock_communicate.return_value = ("", "")

        self.mock_popen.communicate = self.mock_communicate
        self.mock_popen.returncode = 0

        self.mock_subprocess.Popen.return_value = self.mock_popen
        self.mock_subprocess.PIPE = mock.Mock()

        self.mock_utils = mock.Mock()
        self.mock_utils.ensure_dir_exists = mock.Mock()

        self.run_oscap_remediate = common.run_oscap_remediate
        self.run_oscap_remediate.func_globals["subprocess"] = self.mock_subprocess
        self.run_oscap_remediate.func_globals["utils"] = self.mock_utils

    def run_oscap_remediate_profile_only_test(self):
        self.run_oscap_remediate("myprofile", "my_ds.xml")

        # check calls where done right
        args = ["oscap", "xccdf", "eval", "--remediate",
                "--results=%s" % common.RESULTS_PATH,
                "--report=%s" % common.REPORT_PATH,
                "--profile=myprofile",
                "my_ds.xml"]

        # it's impossible to check the preexec_func as it is an internal
        # function of the run_oscap_remediate function
        kwargs = {"stdout": self.mock_subprocess.PIPE,
                  "stderr": self.mock_subprocess.PIPE,
                  }

        for arg in args:
            self.assertIn(arg, self.mock_subprocess.Popen.call_args[0][0])
            self.mock_subprocess.Popen.call_args[0][0].remove(arg)

        # nothing else should have been passed
        self.assertEqual(self.mock_subprocess.Popen.call_args[0][0], [])

        for (key, val) in kwargs.iteritems():
            self.assertEqual(kwargs[key],
                             self.mock_subprocess.Popen.call_args[1].pop(key))

        # plus the preexec_fn kwarg should have been passed
        self.assertIn("preexec_fn", self.mock_subprocess.Popen.call_args[1])

    def run_oscap_remediate_with_ds_test(self):
        self.run_oscap_remediate("myprofile", "my_ds.xml", "my_ds_id")

        # check calls where done right
        args = ["oscap", "xccdf", "eval", "--remediate",
                "--results=%s" % common.RESULTS_PATH,
                "--report=%s" % common.REPORT_PATH,
                "--profile=myprofile",
                "--datastream-id=my_ds_id", "my_ds.xml"]

        # it's impossible to check the preexec_func as it is an internal
        # function of the run_oscap_remediate function
        kwargs = {"stdout": self.mock_subprocess.PIPE,
                  "stderr": self.mock_subprocess.PIPE,
                  }

        for arg in args:
            self.assertIn(arg, self.mock_subprocess.Popen.call_args[0][0])
            self.mock_subprocess.Popen.call_args[0][0].remove(arg)

        # nothing else should have been passed
        self.assertEqual(self.mock_subprocess.Popen.call_args[0][0], [])

        for (key, val) in kwargs.iteritems():
            self.assertEqual(kwargs[key],
                             self.mock_subprocess.Popen.call_args[1].pop(key))

        # plus the preexec_fn kwarg should have been passed
        self.assertIn("preexec_fn", self.mock_subprocess.Popen.call_args[1])

    def run_oscap_remediate_with_ds_xccdf_test(self):
        self.run_oscap_remediate("myprofile", "my_ds.xml", "my_ds_id",
                                 "my_xccdf_id")

        # check calls where done right
        args = ["oscap", "xccdf", "eval", "--remediate",
                "--results=%s" % common.RESULTS_PATH,
                "--report=%s" % common.REPORT_PATH,
                "--profile=myprofile",
                "--datastream-id=my_ds_id", "--xccdf-id=my_xccdf_id",
                "my_ds.xml"]

        # it's impossible to check the preexec_func as it is an internal
        # function of the run_oscap_remediate function
        kwargs = {"stdout": self.mock_subprocess.PIPE,
                  "stderr": self.mock_subprocess.PIPE,
                  }

        for arg in args:
            self.assertIn(arg, self.mock_subprocess.Popen.call_args[0][0])
            self.mock_subprocess.Popen.call_args[0][0].remove(arg)

        # nothing else should have been passed
        self.assertEqual(self.mock_subprocess.Popen.call_args[0][0], [])

        for (key, val) in kwargs.iteritems():
            self.assertEqual(kwargs[key],
                             self.mock_subprocess.Popen.call_args[1].pop(key))

        # plus the preexec_fn kwarg should have been passed
        self.assertIn("preexec_fn", self.mock_subprocess.Popen.call_args[1])

    def run_oscap_remediate_create_dir_test(self):
        self.run_oscap_remediate("myprofile", "my_ds.xml")

        self.mock_utils.ensure_dir_exists.assert_called_with(
            os.path.dirname(common.RESULTS_PATH))

    def run_oscap_remediate_create_chroot_dir_test(self):
        self.run_oscap_remediate("myprofile", "my_ds.xml", chroot="/mnt/test")

        chroot_dir = "/mnt/test" + os.path.dirname(common.RESULTS_PATH)
        self.mock_utils.ensure_dir_exists.assert_called_with(chroot_dir)


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2013  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Vratislav Podzimek <vpodzime@redhat.com>
#

"""
Module with various classes for SCAP content processing and retrieving data from
it.

"""

import os.path

from collections import namedtuple, OrderedDict
from openscap_api import OSCAP
from pyanaconda.iutil import execReadlines
try:
    from html.parser import HTMLParser
except ImportError:
    from HTMLParser import HTMLParser


class ContentHandlingError(Exception):
    """Exception class for errors related to SCAP content handling."""

    pass


class DataStreamHandlingError(ContentHandlingError):
    """Exception class for errors related to data stream handling."""

    pass


class BenchmarkHandlingError(ContentHandlingError):
    """Exception class for errors related to benchmark handling."""

    pass


class ContentCheckError(ContentHandlingError):
    """Exception class for errors related to content (integrity,...) checking."""

    pass


class ParseHTMLContent(HTMLParser):
    """Parser class for HTML tags within content"""

    def __init__(self):
        HTMLParser.__init__(self)
        self.content = ""

    def handle_starttag(self, tag, attrs):
        if tag == "html:ul":
            self.content += "\n"
        elif tag == "html:li":
            self.content += "\n"
        elif tag == "html:br":
            self.content += "\n"

    def handle_endtag(self, tag):
        if tag == "html:ul":
            self.content += "\n"
        elif tag == "html:li":
            self.content += "\n"

    def handle_data(self, data):
        self.content += data.strip()

    def get_content(self):
        return self.content


def parse_HTML_from_content(content):
    """This is a very simple HTML to text parser.

    HTML tags will be removed while trying to maintain readability
    of content.

    :param content: content whose HTML tags will be parsed
    :return: content without HTML tags
    """

    parser = ParseHTMLContent()
    parser.feed(content)
    return parser.get_content()


# namedtuple class (not a constant, pylint!) for info about a XCCDF profile
# pylint: disable-msg=C0103
ProfileInfo = namedtuple("ProfileInfo", ["id", "title", "description"])

# namedtuple class for info about content files found
# pylint: disable-msg=C0103
ContentFiles = namedtuple("ContentFiles", ["xccdf", "cpe", "tailoring"])


def oscap_text_itr_get_text(itr):
    """
    Helper function for getting a text from the oscap_text_iterator.

    :param itr: oscap_text_iterator to get the text from
    :type itr: oscap_text_iterator
    :return: text gotten from the iterator
    :rtype: str

    """

    ret = ""
    while OSCAP.oscap_text_iterator_has_more(itr):
        text_item = OSCAP.oscap_text_iterator_next(itr)
        ret += OSCAP.oscap_text_get_text(text_item)

    return ret


def explore_content_files(fpaths):
    """
    Function for finding content files in a list of file paths. SIMPLY PICKS
    THE FIRST USABLE CONTENT FILE OF A PARTICULAR TYPE AND JUST PREFERS DATA
    STREAMS OVER STANDALONE BENCHMARKS.

    :param fpaths: a list of file paths to search for content files in
    :type fpaths: [str]
    :return: a tuple containing the content handling class and an ContentFiles
             instance containing the file names of the XCCDF file, CPE
             dictionary and tailoring file or "" in place of those items if not
             found
    :rtype: (class, ContentFiles)

    """

    def get_doc_type(file_path):
        try:
            for line in execReadlines("oscap", ["info", file_path]):
                if line.startswith("Document type:"):
                    _prefix, _sep, type_info = line.partition(":")
                    return type_info.strip()
        except OSError:
            # 'oscap info' exitted with a non-zero exit code -> unknown doc type
            return None

    xccdf_file = ""
    cpe_file = ""
    tailoring_file = ""
    found_ds = False
    content_class = None

    for fpath in fpaths:
        doc_type = get_doc_type(fpath)
        if not doc_type:
            continue

        # prefer DS over standalone XCCDF
        if doc_type == "Source Data Stream" and (not xccdf_file or not found_ds):
            xccdf_file = fpath
            content_class = DataStreamHandler
            found_ds = True
        elif doc_type == "XCCDF Checklist" and not xccdf_file:
            xccdf_file = fpath
            content_class = BenchmarkHandler
        elif doc_type == "CPE Dictionary" and not cpe_file:
            cpe_file = fpath
        elif doc_type == "XCCDF Tailoring" and not tailoring_file:
            tailoring_file = fpath

    # TODO: raise exception if no xccdf_file is found?
    files = ContentFiles(xccdf_file, cpe_file, tailoring_file)
    return (content_class, files)


class DataStreamHandler(object):
    """
    Class for handling data streams in the data stream collection and
    retrieving data from it. For example a list of data stream indices,
    checklists in a given data stream of profiles.

    """

    def __init__(self, dsc_file_path, tailoring_file_path=""):
        """
        Constructor for the DataStreamHandler class.

        :param dsc_file_path: path to a file with a data stream collection
        :type dsc_file_path: str
        :param tailoring_file_path: path to a tailoring file
        :type tailoring_file_path: str

        """

        # is used to speed up getting lists of profiles
        self._profiles_cache = dict()

        if not os.path.exists(dsc_file_path):
            msg = "Invalid file path: '%s'" % dsc_file_path
            raise DataStreamHandlingError(msg)

        self._dsc_file_path = dsc_file_path

        # create an XCCDF session for the file
        self._session = OSCAP.xccdf_session_new(dsc_file_path)
        if not self._session:
            msg = "'%s' is not a valid SCAP content file" % dsc_file_path
            raise DataStreamHandlingError(msg)
        if OSCAP.xccdf_session_load(self._session) != 0:
            raise DataStreamHandlingError(OSCAP.oscap_err_desc())

        if tailoring_file_path:
            OSCAP.xccdf_session_set_user_tailoring_file(self._session,
                                                        tailoring_file_path)

        if not OSCAP.xccdf_session_is_sds(self._session):
            msg = "'%s' is not a data stream collection" % dsc_file_path
            raise DataStreamHandlingError(msg)

        # dictionary holding the items gathered from DSC processing
        self._items = OrderedDict()

        # create an sds index for the content
        self._sds_idx = OSCAP.xccdf_session_get_sds_idx(self._session)

        # iterate over streams and get checklists from each stream
        streams_itr = OSCAP.ds_sds_index_get_streams(self._sds_idx)
        while OSCAP.ds_stream_index_iterator_has_more(streams_itr):
            stream_idx = OSCAP.ds_stream_index_iterator_next(streams_itr)

            # will be used to store the checklists for streams
            stream_id = OSCAP.ds_stream_index_get_id(stream_idx)
            checklists = []

            # iterate over checklists and append their ids to the list
            chklist_itr = OSCAP.ds_stream_index_get_checklists(stream_idx)
            while OSCAP.oscap_string_iterator_has_more(chklist_itr):
                checklists.append(OSCAP.oscap_string_iterator_next(chklist_itr))

            # store the list of checklist for the current stream
            self._items[stream_id] = checklists

            OSCAP.oscap_string_iterator_free(chklist_itr)

        OSCAP.ds_stream_index_iterator_free(streams_itr)

    def __del__(self):
        """Destructor for the DataStreamHandler class."""

        if '_session' in locals():
            # we should free the session
            OSCAP.xccdf_session_free(self._session)

    def get_data_streams(self):
        """
        Method to get a list of data streams found in the data stream
        collection.

        :return: list of data stream IDs
        :rtype: list of strings

        """

        return self._items.keys()

    def get_data_streams_checklists(self):
        """
        Method to get data streams and their checklists found in the data
        stream collection.

        :return: a dictionary consisting of the IDs of the data streams as keys
                 and lists of their checklists' IDs as values
        :rtype: dict(str -> list of strings)

        """

        # easy, we already have exactly what should be returned, just create a
        # copy, so that the caller cannot modify our internal attributes
        return dict(self._items)

    def get_checklists(self, data_stream_id):
        """
        Method to get a list of checklists found in the data stream given by
        the data_stream_id.

        :param data_stream_id: ID of the data stream to get checklists from
        :type data_stream_id: str
        :return: list of checklist IDs found in the data stream given by the ID
        :rtype: list of strings

        """

        if data_stream_id not in self._items:
            msg = "Invalid data stream id given: '%s'" % data_stream_id
            raise DataStreamHandlingError(msg)

        return self._items[data_stream_id]

    def get_profiles(self, data_stream_id, checklist_id):
        """
        Method to get a list of profiles defined in the checklist given by the
        checklist_id that is defined in the data stream given by the
        data_stream_id.

        :param data_stream_id: ID of the data stream to get checklists from
        :type data_stream_id: str
        :param checklist_id: ID of the checklist to get profiles from
        :type checklist_id: str
        :return: list of profiles found in the checklist
        :rtype: list of ProfileInfo instances

        """

        cache_id = "%s;%s" % (data_stream_id, checklist_id)
        if cache_id in self._profiles_cache:
            # found in cache, return the value
            return self._profiles_cache[cache_id]

        # not found in the cache, needs to be gathered

        # set the data stream and component (checklist) for the session
        OSCAP.xccdf_session_free(self._session)

        self._session = OSCAP.xccdf_session_new(self._dsc_file_path)
        if not self._session:
            msg = "'%s' is not a valid SCAP content file" % self._dsc_file_path
            raise DataStreamHandlingError(msg)

        OSCAP.xccdf_session_set_datastream_id(self._session, data_stream_id)
        OSCAP.xccdf_session_set_component_id(self._session, checklist_id)
        if OSCAP.xccdf_session_load(self._session) != 0:
            raise DataStreamHandlingError(OSCAP.oscap_err_desc())

        # get the benchmark (checklist)
        policy_model = OSCAP.xccdf_session_get_policy_model(self._session)

        default_policy = OSCAP.xccdf_policy_new(policy_model, None)
        default_rules_count = OSCAP.xccdf_policy_get_selected_rules_count(default_policy)

        # will hold items for the profiles for the speficied DS and checklist
        profiles = []

        if default_rules_count > 0:
            profiles.append(ProfileInfo("default", "Default",
                            "The implicit XCCDF profile. Usually, the default contains no rules."))

        benchmark = OSCAP.xccdf_policy_model_get_benchmark(policy_model)

        # iterate over the profiles in the benchmark and store them
        profile_itr = OSCAP.xccdf_benchmark_get_profiles(benchmark)
        while OSCAP.xccdf_profile_iterator_has_more(profile_itr):
            profile = OSCAP.xccdf_profile_iterator_next(profile_itr)

            id_ = OSCAP.xccdf_profile_get_id(profile)
            title = oscap_text_itr_get_text(OSCAP.xccdf_profile_get_title(profile))
            desc = parse_HTML_from_content(oscap_text_itr_get_text(OSCAP.xccdf_profile_get_description(profile)))
            info = ProfileInfo(id_, title, desc)

            profiles.append(info)

        OSCAP.xccdf_profile_iterator_free(profile_itr)

        # cache the result
        self._profiles_cache[cache_id] = profiles

        return profiles


class BenchmarkHandler(object):
    """
    Class for handling XCCDF benchmark and retrieving data from it (mainly the
    list of profiles).

    """

    def __init__(self, xccdf_file_path, tailoring_file_path=""):
        """
        Constructor for the BenchmarkHandler class.

        :param xccdf_file_path: path to a file with an XCCDF benchmark
        :type xccdf_file_path: str
        :param tailoring_file_path: path to a tailoring file
        :type tailoring_file_path: str
        """

        if not os.path.exists(xccdf_file_path):
            msg = "Invalid file path: '%s'" % xccdf_file_path
            raise BenchmarkHandlingError(msg)

        session = OSCAP.xccdf_session_new(xccdf_file_path)
        if not session:
            msg = "'%s' is not a valid SCAP content file" % xccdf_file_path
            raise BenchmarkHandlingError(msg)

        if tailoring_file_path:
            OSCAP.xccdf_session_set_user_tailoring_file(session,
                                                        tailoring_file_path)
        if OSCAP.xccdf_session_load(session) != 0:
            raise BenchmarkHandlingError(OSCAP.oscap_err_desc())

        # get the benchmark object
        policy_model = OSCAP.xccdf_session_get_policy_model(session)
        benchmark = OSCAP.xccdf_policy_model_get_benchmark(policy_model)

        default_policy = OSCAP.xccdf_policy_new(policy_model, None)
        default_rules_count = OSCAP.xccdf_policy_get_selected_rules_count(default_policy)

        # stores a list of profiles in the benchmark
        self._profiles = []

        if default_rules_count > 0:
            self._profiles.append(ProfileInfo("default", "Default",
                                  "The implicit XCCDF profile. Usually, the default contains no rules."))

        if not benchmark:
            msg = "Not a valid benchmark file: '%s'" % xccdf_file_path
            raise BenchmarkHandlingError(msg)

        # iterate over the profiles in the benchmark and store them
        profile_itr = OSCAP.xccdf_benchmark_get_profiles(benchmark)
        while OSCAP.xccdf_profile_iterator_has_more(profile_itr):
            profile = OSCAP.xccdf_profile_iterator_next(profile_itr)

            id_ = OSCAP.xccdf_profile_get_id(profile)
            title = oscap_text_itr_get_text(OSCAP.xccdf_profile_get_title(profile))
            desc = parse_HTML_from_content(oscap_text_itr_get_text(OSCAP.xccdf_profile_get_description(profile)))
            info = ProfileInfo(id_, title, desc)

            self._profiles.append(info)

        if tailoring_file_path:
            tailoring = OSCAP.xccdf_policy_model_get_tailoring(policy_model)
            profile_itr = OSCAP.xccdf_tailoring_get_profiles(tailoring)
            while OSCAP.xccdf_profile_iterator_has_more(profile_itr):
                profile = OSCAP.xccdf_profile_iterator_next(profile_itr)

                id_ = OSCAP.xccdf_profile_get_id(profile)
                title = oscap_text_itr_get_text(OSCAP.xccdf_profile_get_title(profile))
                desc = parse_HTML_from_content(oscap_text_itr_get_text(OSCAP.xccdf_profile_get_description(profile)))
                info = ProfileInfo(id_, title, desc)

                self._profiles.append(info)

        OSCAP.xccdf_profile_iterator_free(profile_itr)
        OSCAP.xccdf_session_free(session)

    @property
    def profiles(self):
        """Property for the list of profiles defined in the benchmark."""

        return self._profiles



#
# Copyright (C) 2013  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Vratislav Podzimek <vpodzime@redhat.com>
#

"""Module with the OSCAPdata class."""

import shutil
import re
import os
import time
import logging
import gettext

from pyanaconda.addons import AddonData
from pyanaconda.iutil import getSysroot
from pyanaconda.progress import progressQ
from pyanaconda import errors
from pyanaconda import iutil
from pyanaconda import flags
from pykickstart.errors import KickstartParseError, KickstartValueError
from org_fedora_oscap import utils, common, rule_handling, data_fetch
from org_fedora_oscap.common import SUPPORTED_ARCHIVES
from org_fedora_oscap.content_handling import ContentCheckError

log = logging.getLogger("anaconda")

_ = lambda x: gettext.ldgettext("oscap-anaconda-addon", x)

# export OSCAPdata class to prevent Anaconda's collect method from taking
# AddonData class instead of the OSCAPdata class
# @see: pyanaconda.kickstart.AnacondaKSHandler.__init__
__all__ = ["OSCAPdata"]

SUPPORTED_CONTENT_TYPES = ("datastream", "rpm", "archive",
                           "scap-security-guide",
                           )

SUPPORTED_URL_PREFIXES = ("http://", "https://", "ftp://"
                          # LABEL:?, hdaX:?,
                          )

REQUIRED_PACKAGES = ("openscap", "openscap-scanner", )

FINGERPRINT_REGEX = re.compile(r'^[a-z0-9]+$')


class MisconfigurationError(common.OSCAPaddonError):
    """Exception for reporting misconfiguration."""

    pass


class OSCAPdata(AddonData):
    """
    Class parsing and storing data for the OSCAP addon.

    :see: pyanaconda.addons.AddonData

    """

    def __init__(self, name, just_clear=False):
        """
        :param name: name of the addon
        :type name: str

        """

        if not just_clear:
            # do not call the parent's __init__ more than once
            AddonData.__init__(self, name)

        # values specifying the content
        self.content_type = ""
        self.content_url = ""
        self.datastream_id = ""
        self.xccdf_id = ""
        self.profile_id = ""
        self.content_path = ""
        self.cpe_path = ""
        self.tailoring_path = ""

        # additional values
        self.fingerprint = ""

        # certificate to verify HTTPS connection or signed data
        self.certificates = ""

        # internal values
        self.rule_data = rule_handling.RuleData()
        self.dry_run = False

    def __str__(self):
        """
        What should end up in the resulting kickstart file, i.e. string
        representation of the stored data.

        """

        if self.dry_run or not self.profile_id:
            # the addon was run in the dry run mode, omit it from the kickstart
            return ""

        def key_value_pair(key, value, indent=4):
            return "%s%s = %s" % (indent * " ", key, value)

        ret = "%%addon %s" % self.name
        ret += "\n%s" % key_value_pair("content-type", self.content_type)

        if self.content_url:
            ret += "\n%s" % key_value_pair("content-url", self.content_url)
        if self.datastream_id:
            ret += "\n%s" % key_value_pair("datastream-id", self.datastream_id)
        if self.xccdf_id:
            ret += "\n%s" % key_value_pair("xccdf-id", self.xccdf_id)
        if self.content_path and self.content_type != "scap-security-guide":
            ret += "\n%s" % key_value_pair("content-path", self.content_path)
        if self.cpe_path:
            ret += "\n%s" % key_value_pair("cpe-path", self.cpe_path)
        if self.tailoring_path:
            ret += "\n%s" % key_value_pair("tailoring-path",
                                           self.tailoring_path)

        ret += "\n%s" % key_value_pair("profile", self.profile_id)

        if self.fingerprint:
            ret += "\n%s" % key_value_pair("fingerprint", self.fingerprint)

        if self.certificates:
            ret += "\n%s" % key_value_pair("certificates", self.certificates)

        ret += "\n%end\n\n"
        return ret

    def _parse_content_type(self, value):
        value_low = value.lower()
        if value_low in SUPPORTED_CONTENT_TYPES:
            self.content_type = value_low
        else:
            msg = "Unsupported content type '%s' in the %s addon" % (value,
                                                                     self.name)
            raise KickstartValueError(msg)

    def _parse_content_url(self, value):
        if any(value.startswith(prefix)
               for prefix in SUPPORTED_URL_PREFIXES):
            self.content_url = value
        else:
            msg = "Unsupported url '%s' in the %s addon" % (value, self.name)
            raise KickstartValueError(msg)

    def _parse_datastream_id(self, value):
        # need to be checked?
        self.datastream_id = value

    def _parse_xccdf_id(self, value):
        # need to be checked?
        self.xccdf_id = value

    def _parse_profile_id(self, value):
        # need to be checked?
        self.profile_id = value

    def _parse_content_path(self, value):
        # need to be checked?
        self.content_path = value

    def _parse_cpe_path(self, value):
        # need to be checked?
        self.cpe_path = value

    def _parse_tailoring_path(self, value):
        # need to be checked?
        self.tailoring_path = value

    def _parse_fingerprint(self, value):
        if FINGERPRINT_REGEX.match(value) is None:
            msg = "Unsupported or invalid fingerprint"
            raise KickstartValueError(msg)

        if utils.get_hashing_algorithm(value) is None:
            msg = "Unsupported fingerprint"
            raise KickstartValueError(msg)

        self.fingerprint = value

    def _parse_certificates(self, value):
        self.certificates = value

    def handle_line(self, line):
        """
        The handle_line method that is called with every line from this addon's
        %addon section of the kickstart file.

        :param line: a single line from the %addon section
        :type line: str

        """

        actions = {"content-type": self._parse_content_type,
                   "content-url": self._parse_content_url,
                   "content-path": self._parse_content_path,
                   "datastream-id": self._parse_datastream_id,
                   "profile": self._parse_profile_id,
                   "xccdf-id": self._parse_xccdf_id,
                   "xccdf-path": self._parse_content_path,
                   "cpe-path": self._parse_cpe_path,
                   "tailoring-path": self._parse_tailoring_path,
                   "fingerprint": self._parse_fingerprint,
                   "certificates": self._parse_certificates,
                   }

        line = line.strip()
        (pre, sep, post) = line.partition("=")
        pre = pre.strip()
        post = post.strip()
        post = post.strip('"')

        try:
            actions[pre](post)
        except KeyError:
            msg = "Unknown item '%s' for %s addon" % (line, self.name)
            raise KickstartParseError(msg)

    def finalize(self):
        """
        The finalize method that is called when the end of the %addon section
        (the %end line) is reached. It means no more kickstart data will come.

        """

        tmpl = "%s missing for the %s addon"

        # check provided data
        if not self.content_type:
            raise KickstartValueError(tmpl % ("content-type", self.name))

        if self.content_type != "scap-security-guide" and not self.content_url:
            raise KickstartValueError(tmpl % ("content-url", self.name))

        if not self.profile_id:
            self.profile_id = "default"

        if self.content_type in ("rpm", "archive") and not self.content_path:
            msg = "Path to the XCCDF file has to be given if content in RPM "\
                  "or archive is used"
            raise KickstartValueError(msg)

        if self.content_type == "rpm" and not self.content_url.endswith(".rpm"):
            msg = "Content type set to RPM, but the content URL doesn't end "\
                  "with '.rpm'"
            raise KickstartValueError(msg)

        if self.content_type == "archive":
            supported_archive = any(self.content_url.endswith(arch_type)
                                    for arch_type in SUPPORTED_ARCHIVES)
            if not supported_archive:
                msg = "Unsupported archive type of the content "\
                      "file '%s'" % self.content_url
                raise KickstartValueError(msg)

        # do some initialization magic in case of SSG
        if self.content_type == "scap-security-guide":
            if not common.ssg_available():
                msg = "SCAP Security Guide not found on the system"
                raise KickstartValueError(msg)

            self.content_path = common.SSG_DIR + common.SSG_CONTENT

    @property
    def content_defined(self):
        return self.content_url or self.content_type == "scap-security-guide"

    @property
    def content_name(self):
        if self.content_type == "scap-security-guide":
            raise ValueError("Using scap-security-guide, no single content file")

        rest = "/anonymous_content"
        for prefix in SUPPORTED_URL_PREFIXES:
            if self.content_url.startswith(prefix):
                rest = self.content_url[len(prefix):]
                break

        parts = rest.rsplit("/", 1)
        if len(parts) != 2:
            msg = "Unsupported url '%s' in the %s addon" % (self.content_url,
                                                            self.name)
            raise KickstartValueError(msg)

        return parts[1]

    @property
    def raw_preinst_content_path(self):
        """Path to the raw (unextracted, ...) pre-installation content file"""

        return utils.join_paths(common.INSTALLATION_CONTENT_DIR,
                                self.content_name)

    @property
    def raw_postinst_content_path(self):
        """Path to the raw (unextracted, ...) post-installation content file"""

        return utils.join_paths(common.TARGET_CONTENT_DIR,
                                self.content_name)

    @property
    def preinst_content_path(self):
        """Path to the pre-installation content file"""

        if self.content_type == "datastream":
            return utils.join_paths(common.INSTALLATION_CONTENT_DIR,
                                    self.content_name)
        elif self.content_type == "scap-security-guide":
            # SSG is not copied to the standard place
            return self.content_path
        else:
            return utils.join_paths(common.INSTALLATION_CONTENT_DIR,
                                    self.content_path)

    @property
    def postinst_content_path(self):
        """Path to the post-installation content file"""

        if self.content_type == "datastream":
            return utils.join_paths(common.TARGET_CONTENT_DIR,
                                    self.content_name)
        elif self.content_type in ("rpm", "scap-security-guide"):
            # no path magic in case of RPM (SSG is installed as an RPM)
            return self.content_path
        else:
            return utils.join_paths(common.TARGET_CONTENT_DIR,
                                    self.content_path)

    @property
    def preinst_tailoring_path(self):
        """Path to the pre-installation tailoring file (if any)"""

        if not self.tailoring_path:
            return ""

        return utils.join_paths(common.INSTALLATION_CONTENT_DIR,
                                self.tailoring_path)

    @property
    def postinst_tailoring_path(self):
        """Path to the post-installation tailoring file (if any)"""

        if not self.tailoring_path:
            return ""

        if self.content_type == "rpm":
            # no path magic in case of RPM
            return self.tailoring_path

        return utils.join_paths(common.TARGET_CONTENT_DIR,
                                self.tailoring_path)

    def _fetch_content_and_initialize(self):
        """Fetch content and initialize from it"""

        data_fetch.fetch_data(self.content_url, self.raw_preinst_content_path,
                              self.certificates)
        # RPM is an archive at this phase
        if self.content_type in ("archive", "rpm"):
            # extract the content
            common.extract_data(self.raw_preinst_content_path,
                                common.INSTALLATION_CONTENT_DIR,
                                [self.content_path])

        rules = common.get_fix_rules_pre(self.profile_id,
                                         self.preinst_content_path,
                                         self.datastream_id, self.xccdf_id,
                                         self.preinst_tailoring_path)

        # parse and store rules with a clean RuleData instance
        self.rule_data = rule_handling.RuleData()
        for rule in rules.splitlines():
            self.rule_data.new_rule(rule)

    def setup(self, storage, ksdata, instclass, payload):
        """
        The setup method that should make changes to the runtime environment
        according to the data stored in this object.

        :param storage: object storing storage-related information
                        (disks, partitioning, bootloader, etc.)
        :type storage: blivet.Blivet instance
        :param ksdata: data parsed from the kickstart file and set in the
                       installation process
        :type ksdata: pykickstart.base.BaseHandler instance
        :param instclass: distribution-specific information
        :type instclass: pyanaconda.installclass.BaseInstallClass

        """

        if self.dry_run or not self.profile_id:
            # nothing more to be done in the dry-run mode or if no profile is
            # selected
            return

        if not os.path.exists(self.preinst_content_path) and not os.path.exists(self.raw_preinst_content_path):
            # content not available/fetched yet
            try:
                self._fetch_content_and_initialize()
            except (common.OSCAPaddonError, data_fetch.DataFetchError) as e:
                log.error("Failed to fetch and initialize SCAP content!")
                msg = _("There was an error fetching and loading the security content:\n" +
                        "%s\n" +
                        "The installation should be aborted. Do you wish to continue anyway?") % e

                if flags.flags.automatedInstall and not flags.flags.ksprompt:
                    # cannot have ask in a non-interactive kickstart
                    # installation
                    raise errors.CmdlineError(msg)

                answ = errors.errorHandler.ui.showYesNoQuestion(msg)
                if answ == errors.ERROR_CONTINUE:
                    # prevent any futher actions here by switching to the dry
                    # run mode and let things go on
                    self.dry_run = True
                    return
                else:
                    # Let's sleep forever to prevent any further actions and
                    # wait for the main thread to quit the process.
                    progressQ.send_quit(1)
                    while True:
                        time.sleep(100000)

        # check fingerprint if given
        if self.fingerprint:
            hash_obj = utils.get_hashing_algorithm(self.fingerprint)
            digest = utils.get_file_fingerprint(self.raw_preinst_content_path,
                                                hash_obj)
            if digest != self.fingerprint:
                log.error("Failed to fetch and initialize SCAP content!")
                msg = _("The integrity check of the security content failed.\n" +
                        "The installation should be aborted. Do you wish to continue anyway?")

                if flags.flags.automatedInstall and not flags.flags.ksprompt:
                    # cannot have ask in a non-interactive kickstart
                    # installation
                    raise errors.CmdlineError(msg)

                answ = errors.errorHandler.ui.showYesNoQuestion(msg)
                if answ == errors.ERROR_CONTINUE:
                    # prevent any futher actions here by switching to the dry
                    # run mode and let things go on
                    self.dry_run = True
                    return
                else:
                    # Let's sleep forever to prevent any further actions and
                    # wait for the main thread to quit the process.
                    progressQ.send_quit(1)
                    while True:
                        time.sleep(100000)

        # evaluate rules, do automatic fixes and stop if something that cannot
        # be fixed automatically is wrong
        fatal_messages = [message for message in self.rule_data.eval_rules(ksdata, storage)
                          if message.type == common.MESSAGE_TYPE_FATAL]
        if any(fatal_messages):
            msg = "Wrong configuration detected!\n"
            msg += "\n".join(message.text for message in fatal_messages)
            msg += "\nThe installation should be aborted. Do you wish to continue anyway?"
            if flags.flags.automatedInstall and not flags.flags.ksprompt:
                # cannot have ask in a non-interactive kickstart installation
                raise errors.CmdlineError(msg)

            answ = errors.errorHandler.ui.showYesNoQuestion(msg)
            if answ == errors.ERROR_CONTINUE:
                # prevent any futher actions here by switching to the dry
                # run mode and let things go on
                self.dry_run = True
                return
            else:
                # Let's sleep forever to prevent any further actions and wait
                # for the main thread to quit the process.
                progressQ.send_quit(1)
                while True:
                    time.sleep(100000)

        # add packages needed on the target system to the list of packages
        # that are requested to be installed
        pkgs_to_install = list(REQUIRED_PACKAGES)
        if self.content_type == "scap-security-guide":
            pkgs_to_install.append("scap-security-guide")
        for pkg in pkgs_to_install:
            if pkg not in ksdata.packages.packageList:
                ksdata.packages.packageList.append(pkg)

    def execute(self, storage, ksdata, instclass, users, payload):
        """
        The execute method that should make changes to the installed system. It
        is called only once in the post-install setup phase.

        :see: setup
        :param users: information about created users
        :type users: pyanaconda.users.Users instance

        """

        if self.dry_run or not self.profile_id:
            # nothing more to be done in the dry-run mode or if no profile is
            # selected
            return

        target_content_dir = utils.join_paths(getSysroot(),
                                              common.TARGET_CONTENT_DIR)
        utils.ensure_dir_exists(target_content_dir)

        if self.content_type == "datastream":
            shutil.copy2(self.preinst_content_path, target_content_dir)
        elif self.content_type == "rpm":
            # copy the RPM to the target system
            shutil.copy2(self.raw_preinst_content_path, target_content_dir)

            # and install it with yum
            ret = iutil.execInSysroot("yum", ["-y", "--nogpg", "install",
                                              self.raw_postinst_content_path])
            if ret != 0:
                raise common.ExtractionError("Failed to install content "
                                             "RPM to the target system")
        elif self.content_type == "scap-security-guide":
            # nothing needed
            pass
        else:
            utils.universal_copy(utils.join_paths(common.INSTALLATION_CONTENT_DIR,
                                                  "*"),
                                 target_content_dir)
        if os.path.exists(self.preinst_tailoring_path):
            shutil.copy2(self.preinst_tailoring_path, target_content_dir)

        common.run_oscap_remediate(self.profile_id, self.postinst_content_path,
                                   self.datastream_id, self.xccdf_id,
                                   self.postinst_tailoring_path,
                                   chroot=getSysroot())

    def clear_all(self):
        """Clear all the stored values."""

        self.__init__(self.name, just_clear=True)



"""Module with tests for the ks/oscap.py module."""

import unittest
import os
from pykickstart.errors import KickstartValueError
from org_fedora_oscap.ks.oscap import OSCAPdata
from org_fedora_oscap import common


class ParsingTest(unittest.TestCase):
    def setUp(self):
        self.oscap_data = OSCAPdata("org_fedora_oscap")
        for line in ["content-type = datastream\n",
                     "content-url = \"https://example.com/hardening.xml\"\n",
                     "datastream-id = id_datastream_1\n",
                     "xccdf-id = id_xccdf_new\n",
                     "content-path = /usr/share/oscap/testing_ds.xml",
                     "cpe-path = /usr/share/oscap/cpe.xml",
                     "tailoring-path = /usr/share/oscap/tailoring.xml",
                     "profile = \"Web Server\"\n",
                     ]:
            self.oscap_data.handle_line(line)

    def parsing_test(self):
        self.assertEqual(self.oscap_data.content_type, "datastream")
        self.assertEqual(self.oscap_data.content_url,
                         "https://example.com/hardening.xml")
        self.assertEqual(self.oscap_data.datastream_id, "id_datastream_1")
        self.assertEqual(self.oscap_data.xccdf_id, "id_xccdf_new")
        self.assertEqual(self.oscap_data.content_path,
                         "/usr/share/oscap/testing_ds.xml")
        self.assertEqual(self.oscap_data.cpe_path, "/usr/share/oscap/cpe.xml")
        self.assertEqual(self.oscap_data.profile_id, "Web Server")
        self.assertEqual(self.oscap_data.content_name, "hardening.xml")
        self.assertEqual(self.oscap_data.tailoring_path,
                         "/usr/share/oscap/tailoring.xml")

    def properties_test(self):
        self.assertEqual(self.oscap_data.preinst_content_path,
                         common.INSTALLATION_CONTENT_DIR +
                         self.oscap_data.content_name)

        self.assertEqual(self.oscap_data.postinst_content_path,
                         common.TARGET_CONTENT_DIR +
                         self.oscap_data.content_name)

        self.assertEqual(self.oscap_data.raw_preinst_content_path,
                         common.INSTALLATION_CONTENT_DIR +
                         self.oscap_data.content_name)

        self.assertEqual(self.oscap_data.preinst_tailoring_path,
                         os.path.normpath(common.INSTALLATION_CONTENT_DIR +
                                          self.oscap_data.tailoring_path))

        self.assertEqual(self.oscap_data.postinst_tailoring_path,
                         os.path.normpath(common.TARGET_CONTENT_DIR +
                                          self.oscap_data.tailoring_path))

    def str_test(self):
        str_ret = str(self.oscap_data)
        self.assertEqual(str_ret,
                         "%addon org_fedora_oscap\n"
                         "    content-type = datastream\n"
                         "    content-url = https://example.com/hardening.xml\n"
                         "    datastream-id = id_datastream_1\n"
                         "    xccdf-id = id_xccdf_new\n"
                         "    content-path = /usr/share/oscap/testing_ds.xml\n"
                         "    cpe-path = /usr/share/oscap/cpe.xml\n"
                         "    tailoring-path = /usr/share/oscap/tailoring.xml\n"
                         "    profile = Web Server\n"
                         "%end\n\n"
                         )

    def str_parse_test(self):
        self.oscap_data2 = OSCAPdata("org_fedora_oscap")
        str_ret = str(self.oscap_data)
        for line in str_ret.splitlines()[1:-1]:
            if "%end" not in line:
                self.oscap_data2.handle_line(line)

        str_ret2 = str(self.oscap_data)
        self.assertEqual(str_ret, str_ret2)


class BackwardCompatibilityParsingTest(unittest.TestCase):
    def setUp(self):
        self.oscap_data = OSCAPdata("org_fedora_oscap")
        for line in ["content-type = datastream\n",
                     "content-url = \"https://example.com/hardening.xml\"\n",
                     "datastream-id = id_datastream_1\n",
                     "xccdf-id = id_xccdf_new\n",
                     "xccdf-path = /usr/share/oscap/xccdf.xml",
                     "cpe-path = /usr/share/oscap/cpe.xml",
                     "tailoring-path = /usr/share/oscap/tailoring.xml",
                     "profile = \"Web Server\"\n",
                     ]:
            self.oscap_data.handle_line(line)

    def str_test(self):
        str_ret = str(self.oscap_data)
        self.assertEqual(str_ret,
                         "%addon org_fedora_oscap\n"
                         "    content-type = datastream\n"
                         "    content-url = https://example.com/hardening.xml\n"
                         "    datastream-id = id_datastream_1\n"
                         "    xccdf-id = id_xccdf_new\n"
                         "    content-path = /usr/share/oscap/xccdf.xml\n"
                         "    cpe-path = /usr/share/oscap/cpe.xml\n"
                         "    tailoring-path = /usr/share/oscap/tailoring.xml\n"
                         "    profile = Web Server\n"
                         "%end\n\n"
                         )


class IncompleteDataTest(unittest.TestCase):
    def setUp(self):
        self.oscap_data = OSCAPdata("org_fedora_oscap")

    def nothing_given_test(self):
        with self.assertRaises(KickstartValueError):
            self.oscap_data.finalize()

    def no_content_type_test(self):
        for line in ["content-url = http://example.com/test_ds.xml",
                     "profile = Web Server",
                     ]:
            self.oscap_data.handle_line(line)
        with self.assertRaises(KickstartValueError):
            self.oscap_data.finalize()

    def no_content_url_test(self):
        for line in ["content-type = datastream",
                     "profile = Web Server",
                     ]:
            self.oscap_data.handle_line(line)

        with self.assertRaises(KickstartValueError):
            self.oscap_data.finalize()

    def no_profile_test(self):
        for line in ["content-url = http://example.com/test_ds.xml",
                     "content-type = datastream",
                     ]:
            self.oscap_data.handle_line(line)

        self.oscap_data.finalize()
        self.assertEqual(self.oscap_data.profile_id, "default")


class InvalidDataTest(unittest.TestCase):
    def setUp(self):
        self.oscap_data = OSCAPdata("org_fedora_oscap")

    def rpm_without_path_test(self):
        for line in ["content-url = http://example.com/oscap_content.rpm",
                     "content-type = RPM",
                     "profile = Web Server",
                     ]:
            self.oscap_data.handle_line(line)

        with self.assertRaises(KickstartValueError):
            self.oscap_data.finalize()

    def rpm_with_wrong_suffix_test(self):
        for line in ["content-url = http://example.com/oscap_content.xml",
                     "content-type = RPM",
                     "profile = Web Server",
                     ]:
            self.oscap_data.handle_line(line)

        with self.assertRaises(KickstartValueError):
            self.oscap_data.finalize()

    def archive_without_path_test(self):
        for line in ["content-url = http://example.com/oscap_content.tar",
                     "content-type = archive",
                     "profile = Web Server",
                     ]:
            self.oscap_data.handle_line(line)

        with self.assertRaises(KickstartValueError):
            self.oscap_data.finalize()

    def unsupported_archive_type_test(self):
        for line in ["content-url = http://example.com/oscap_content.tbz",
                     "content-type = archive",
                     "profile = Web Server",
                     "xccdf-path = xccdf.xml"
                     ]:
            self.oscap_data.handle_line(line)

        with self.assertRaises(KickstartValueError):
            self.oscap_data.finalize()


class EverythingOKtest(unittest.TestCase):
    def setUp(self):
        self.oscap_data = OSCAPdata("org_fedora_oscap")

    def enough_for_ds_test(self):
        for line in ["content-url = http://example.com/test_ds.xml",
                     "content-type = datastream",
                     "profile = Web Server",
                     ]:
            self.oscap_data.handle_line(line)

        self.oscap_data.finalize()

    def enough_for_rpm_test(self):
        for line in ["content-url = http://example.com/oscap_content.rpm",
                     "content-type = RPM",
                     "profile = Web Server",
                     "xccdf-path = /usr/share/oscap/xccdf.xml"
                     ]:
            self.oscap_data.handle_line(line)

        self.oscap_data.finalize()

    def enough_for_archive_test(self):
        for line in ["content-url = http://example.com/oscap_content.tar",
                     "content-type = archive",
                     "profile = Web Server",
                     "xccdf-path = /usr/share/oscap/xccdf.xml"
                     ]:
            self.oscap_data.handle_line(line)

        self.oscap_data.finalize()


class ArchiveHandlingTest(unittest.TestCase):
    """Tests for handling archives."""

    def setUp(self):
        self.oscap_data = OSCAPdata("org_fedora_oscap")

    def archive_preinst_content_path_test(self):
        for line in ["content-url = http://example.com/oscap_content.tar",
                     "content-type = archive",
                     "profile = Web Server",
                     "xccdf-path = oscap/xccdf.xml"
                     ]:
            self.oscap_data.handle_line(line)

        self.oscap_data.finalize()

        # content_name should be the archive's name
        self.assertEqual(self.oscap_data.content_name, "oscap_content.tar")

        # content path should end with the xccdf path
        self.assertTrue(self.oscap_data.preinst_content_path.endswith(
                                                         "oscap/xccdf.xml"))

    def ds_preinst_content_path_test(self):
        for line in ["content-url = http://example.com/scap_content.xml",
                     "content-type = datastream",
                     "profile = Web Server",
                     ]:
            self.oscap_data.handle_line(line)

        self.oscap_data.finalize()

        # both content_name and content path should point to the data stream
        # XML
        self.assertEqual(self.oscap_data.content_name, "scap_content.xml")
        self.assertTrue(self.oscap_data.preinst_content_path.endswith(
                                                         "scap_content.xml"))

    def archive_raw_content_paths_test(self):
        for line in ["content-url = http://example.com/oscap_content.tar",
                     "content-type = archive",
                     "profile = Web Server",
                     "xccdf-path = oscap/xccdf.xml",
                     "tailoring-path = oscap/tailoring.xml",
                     ]:
            self.oscap_data.handle_line(line)

        self.oscap_data.finalize()

        # content_name should be the archive's name
        self.assertEqual(self.oscap_data.content_name, "oscap_content.tar")

        # content path should end with the archive's name
        self.assertTrue(self.oscap_data.raw_preinst_content_path.endswith(
                                                         "oscap_content.tar"))
        self.assertTrue(self.oscap_data.raw_postinst_content_path.endswith(
                                                         "oscap_content.tar"))

        # tailoring paths should be returned properly
        self.assertEqual(self.oscap_data.preinst_tailoring_path,
                         common.INSTALLATION_CONTENT_DIR +
                         self.oscap_data.tailoring_path)

        self.assertEqual(self.oscap_data.postinst_tailoring_path,
                         common.TARGET_CONTENT_DIR +
                         self.oscap_data.tailoring_path)

    def rpm_raw_content_paths_test(self):
        for line in ["content-url = http://example.com/oscap_content.rpm",
                     "content-type = rpm",
                     "profile = Web Server",
                     "xccdf-path = /usr/share/oscap/xccdf.xml",
                     "tailoring-path = /usr/share/oscap/tailoring.xml",
                     ]:
            self.oscap_data.handle_line(line)

        self.oscap_data.finalize()

        # content_name should be the rpm's name
        self.assertEqual(self.oscap_data.content_name, "oscap_content.rpm")

        # content path should end with the rpm's name
        self.assertTrue(self.oscap_data.raw_preinst_content_path.endswith(
                                                         "oscap_content.rpm"))
        self.assertTrue(self.oscap_data.raw_postinst_content_path.endswith(
                                                         "oscap_content.rpm"))

        # content paths should be returned as expected
        self.assertEqual(self.oscap_data.preinst_content_path,
                         os.path.normpath(common.INSTALLATION_CONTENT_DIR +
                                          self.oscap_data.content_path))

        # when using rpm, content_path doesn't change for the post-installation
        # phase
        self.assertEqual(self.oscap_data.postinst_content_path,
                         self.oscap_data.content_path)

    def ds_raw_content_paths_test(self):
        for line in ["content-url = http://example.com/scap_content.xml",
                     "content-type = datastream",
                     "profile = Web Server",
                     ]:
            self.oscap_data.handle_line(line)

        self.oscap_data.finalize()

        # content_name and content paths should all point to the data stream
        # XML
        self.assertEqual(self.oscap_data.content_name, "scap_content.xml")
        self.assertTrue(self.oscap_data.raw_preinst_content_path.endswith(
                                                         "scap_content.xml"))
        self.assertTrue(self.oscap_data.raw_postinst_content_path.endswith(
                                                         "scap_content.xml"))


class FingerprintTests(unittest.TestCase):
    """Tests for fingerprint pre-processing."""

    def setUp(self):
        self.oscap_data = OSCAPdata("org_fedora_oscap")

    def valid_fingerprints_test(self):
        self.oscap_data.handle_line("fingerprint = %s" % ("a" * 32))
        self.oscap_data.handle_line("fingerprint = %s" % ("a" * 40))
        self.oscap_data.handle_line("fingerprint = %s" % ("a" * 56))
        self.oscap_data.handle_line("fingerprint = %s" % ("a" * 64))
        self.oscap_data.handle_line("fingerprint = %s" % ("a" * 96))
        self.oscap_data.handle_line("fingerprint = %s" % ("a" * 128))

    def invalid_fingerprints_test(self):
        # invalid character
        with self.assertRaisesRegexp(KickstartValueError,
                                     "Unsupported or invalid fingerprint"):
            self.oscap_data.handle_line("fingerprint = %s?" % ("a" * 31))

        # invalid lengths (odd and even)
        with self.assertRaisesRegexp(KickstartValueError,
                                     "Unsupported fingerprint"):
            self.oscap_data.handle_line("fingerprint = %s" % ("a" * 31))
        with self.assertRaisesRegexp(KickstartValueError,
                                     "Unsupported fingerprint"):
            self.oscap_data.handle_line("fingerprint = %s" % ("a" * 41))
        with self.assertRaisesRegexp(KickstartValueError,
                                     "Unsupported fingerprint"):
            self.oscap_data.handle_line("fingerprint = %s" % ("a" * 54))
        with self.assertRaisesRegexp(KickstartValueError,
                                     "Unsupported fingerprint"):
            self.oscap_data.handle_line("fingerprint = %s" % ("a" * 66))
        with self.assertRaisesRegexp(KickstartValueError,
                                     "Unsupported fingerprint"):
            self.oscap_data.handle_line("fingerprint = %s" % ("a" * 98))
        with self.assertRaisesRegexp(KickstartValueError,
                                     "Unsupported fingerprint"):
            self.oscap_data.handle_line("fingerprint = %s" % ("a" * 124))



#
# Copyright (C) 2013  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Vratislav Podzimek <vpodzime@redhat.com>
#

"""Module with various utility functions used by the addon."""

import os
import os.path
import shutil
import glob
import hashlib


def ensure_dir_exists(dirpath):
    """
    Checks if a given directory exists and if not, it creates the directory as
    well as all the nonexisting directories in its path.

    :param dirpath: path to the directory to be checked/created
    :type dirpath: str

    """

    if not dirpath:
        # nothing can be done for an empty string
        return

    if not os.path.isdir(dirpath):
        os.makedirs(dirpath)


def universal_copy(src, dst):
    """
    Function that copies the files or directories specified by the src argument
    to the destination given by the dst argument. It should follow the same
    rules as the standard 'cp' utility.

    :param src: source to copy -- may be a glob, file path or a directory path
    :type src: str
    :param dst: destination to copy to
    :type src: str

    """

    if glob.has_magic(src):
        # src is a glob
        sources = glob.glob(src)
    else:
        # not a glob
        sources = [src]

    for item in sources:
        if os.path.isdir(item):
            if os.path.isdir(dst):
                item = item.rstrip("/")
                dirname = item.rsplit("/", 1)[-1]
                shutil.copytree(item, join_paths(dst, dirname))
            else:
                shutil.copytree(item, dst)
        else:
            shutil.copy2(item, dst)


def keep_type_map(func, iterable):
    """
    Function that maps the given function to items in the given iterable
    keeping the type of the iterable.

    :param func: function to be mapped on the items in the iterable
    :type func: in_item -> out_item
    :param iterable: iterable providing the items the function should be mapped
                     on
    :type iterable: iterable
    :return: iterable providing items produced by the function mapped on the
             input items
    :rtype: the same type as input iterable or generator if the iterable is not
            of any basic Python types

    """

    if isinstance(iterable, dict):
        return dict((func(key), iterable[key]) for key in iterable)

    items_gen = (func(item) for item in iterable)
    if isinstance(iterable, list):
        return list(items_gen)
    elif isinstance(iterable, tuple):
        if iterable.__class__ is tuple:
            return tuple(items_gen)
        else:
            return iterable.__class__(*items_gen)
    elif isinstance(iterable, set):
        return set(items_gen)
    elif isinstance(iterable, str):
        return "".join(items_gen)
    else:
        return items_gen


def join_paths(path1, path2):
    """
    Joins two paths as one would expect -- i.e. just like the os.path.join
    function except for doing crazy things when the second argument is an
    absolute path.

    :param path1: first path
    :type path1: str
    :param path2: second path
    :type path2: str
    :return: path1 and path2 joined with the file separator
    :rtype: str

    """

    # os.path.normpath doesn't squash two starting slashes
    path1.replace("//", "/")

    return os.path.normpath(path1 + os.path.sep + path2)


def get_hashing_algorithm(fingerprint):
    """
    Get hashing algorithm for the given fingerprint or None if fingerprint of
    unsupported length is given.

    :param fingerprint: hexa fingerprint to get the hashing algorithm for
    :type fingerprint: hexadecimal str
    :return: one of the hashlib.* hash objects
    :rtype: hashlib.HASH object

    """

    hashes = (hashlib.md5(), hashlib.sha1(), hashlib.sha224(),
              hashlib.sha256(), hashlib.sha384(), hashlib.sha512())

    if len(fingerprint) % 2 == 1:
        return None

    num_bytes = len(fingerprint) / 2

    for hash_obj in hashes:
        # pylint: disable-msg=E1103
        if hash_obj.digest_size == num_bytes:
            return hash_obj

    return None


def get_file_fingerprint(fpath, hash_obj):
    """
    Get fingerprint of the given file with the given hashing algorithm.

    :param fpath: path to the file to get fingerprint for
    :type fpath: str
    :param hash_obj: hashing algorithm to get fingerprint with
    :type hash_obj: hashlib.HASH
    :return: fingerprint of the given file with the given algorithm
    :rtype: hexadecimal str

    """

    with open(fpath, "r") as fobj:
        bsize = 4*1024
        # process file as 4 KB blocks
        buf = fobj.read(bsize)
        while buf:
            hash_obj.update(buf)
            buf = fobj.read(bsize)

    return hash_obj.hexdigest()



#
# Copyright (C) 2013  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Vratislav Podzimek <vpodzime@redhat.com>
#

"""
Module with various classes and functions needed by the OSCAP addon that are
not specific to any installation mode (tui, gui, ks).

"""

import os
import tempfile
import subprocess
import zipfile
import tarfile
import cpioarchive
import re
import logging

from collections import namedtuple
from functools import wraps
from pyanaconda import constants
from pyanaconda import nm
from pyanaconda.threads import threadMgr, AnacondaThread
from org_fedora_oscap import utils
from org_fedora_oscap.data_fetch import fetch_data

log = logging.getLogger("anaconda")

# everything else should be private
__all__ = ["run_oscap_remediate", "get_fix_rules_pre",
           "wait_and_fetch_net_data", "extract_data", "strip_content_dir",
           "OSCAPaddonError"]

INSTALLATION_CONTENT_DIR = "/tmp/openscap_data/"
TARGET_CONTENT_DIR = "/root/openscap_data/"

SSG_DIR = "/usr/share/xml/scap/ssg/content/"
SSG_CONTENT = "ssg-rhel7-ds.xml"
if constants.shortProductName != 'anaconda':
    if constants.shortProductName == 'fedora':
        SSG_CONTENT  = "ssg-fedora-ds.xml"
    else:
        SSG_CONTENT = "ssg-%s%s-ds.xml" % (constants.shortProductName,
                                            constants.productVersion.strip(".")[0])

RESULTS_PATH = utils.join_paths(TARGET_CONTENT_DIR,
                                "eval_remediate_results.xml")
REPORT_PATH = utils.join_paths(TARGET_CONTENT_DIR,
                               "eval_remediate_report.html")

PRE_INSTALL_FIX_SYSTEM_ATTR = "urn:redhat:anaconda:pre"

THREAD_FETCH_DATA = "AnaOSCAPdataFetchThread"

SUPPORTED_ARCHIVES = (".zip", ".tar", ".tar.gz", ".tar.bz2", )

# buffer size for reading and writing out data (in bytes)
IO_BUF_SIZE = 2 * 1024 * 1024


class OSCAPaddonError(Exception):
    """Exception class for OSCAP addon related errors."""

    pass


class OSCAPaddonNetworkError(OSCAPaddonError):
    """Exception class for OSCAP addon related network errors."""

    pass


class ExtractionError(OSCAPaddonError):
    """Exception class for the extraction errors."""

    pass


MESSAGE_TYPE_FATAL = 0
MESSAGE_TYPE_WARNING = 1
MESSAGE_TYPE_INFO = 2

# namedtuple for messages returned from the rules evaluation
#   origin -- class (inherited from RuleHandler) that generated the message
#   type -- one of the MESSAGE_TYPE_* constants defined above
#   text -- the actual message that should be displayed, logged, ...
RuleMessage = namedtuple("RuleMessage", ["origin", "type", "text"])


def get_fix_rules_pre(profile, fpath, ds_id="", xccdf_id="", tailoring=""):
    """
    Get fix rules for the pre-installation environment for a given profile in a
    given datastream and checklist in a given file.

    :see: run_oscap_remediate
    :see: _run_oscap_gen_fix
    :return: fix rules for a given profile
    :rtype: str

    """

    return _run_oscap_gen_fix(profile, fpath, PRE_INSTALL_FIX_SYSTEM_ATTR,
                              ds_id=ds_id, xccdf_id=xccdf_id,
                              tailoring=tailoring)


def _run_oscap_gen_fix(profile, fpath, template, ds_id="", xccdf_id="",
                       tailoring=""):
    """
    Run oscap tool on a given file to get the contents of fix elements with the
    'system' attribute equal to a given template for a given datastream,
    checklist and profile.

    :see: run_oscap_remediate
    :param template: the value of the 'system' attribute of the fix elements
    :type template: str
    :return: oscap tool's stdout
    :rtype: str

    """

    if not profile:
        return ""

    args = ["oscap", "xccdf", "generate", "fix"]
    args.append("--template=%s" % template)

    # oscap uses the default profile by default
    if profile.lower() != "default":
        args.append("--profile=%s" % profile)
    if ds_id:
        args.append("--datastream-id=%s" % ds_id)
    if xccdf_id:
        args.append("--xccdf-id=%s" % xccdf_id)
    if tailoring:
        args.append("--tailoring-file=%s" % tailoring)

    args.append(fpath)

    try:
        proc = subprocess.Popen(args, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
    except OSError as oserr:
        msg = "Failed to run the oscap tool: %s" % oserr
        raise OSCAPaddonError(msg)

    (stdout, stderr) = proc.communicate()

    messages = re.findall(r'OpenSCAP Error:.*', stderr)
    if messages:
        for message in messages:
            log.warning("OSCAP addon: " + message)

    # pylint thinks Popen has no attribute returncode
    # pylint: disable-msg=E1101
    if proc.returncode != 0:
        msg = "Failed to generate fix rules with the oscap tool: %s" % stderr
        raise OSCAPaddonError(msg)

    return stdout


def run_oscap_remediate(profile, fpath, ds_id="", xccdf_id="", tailoring="",
                        chroot=""):
    """
    Run the evaluation and remediation with the oscap tool on a given file,
    doing the remediation as defined in a given profile defined in a given
    checklist that is a part of a given datastream. If requested, run in
    chroot.

    :param profile: id of the profile that will drive the remediation
    :type profile: str
    :param fpath: path to a file with SCAP content
    :type fpath: str
    :param ds_id: ID of the datastream that contains the checklist defining
                  the profile
    :type ds_id: str
    :param xccdf_id: ID of the checklist that defines the profile
    :type xccdf_id: str
    :param tailoring: path to a tailoring file
    :type tailoring: str
    :param chroot: path to the root the oscap tool should be run in
    :type chroot: str
    :return: oscap tool's stdout (summary of the rules, checks and fixes)
    :rtype: str

    """

    if not profile:
        return ""

    def do_chroot():
        """Helper function doing the chroot if requested."""
        if chroot and chroot != "/":
            os.chroot(chroot)
            os.chdir("/")

    # make sure the directory for the results exists
    results_dir = os.path.dirname(RESULTS_PATH)
    if chroot:
        results_dir = os.path.normpath(chroot + "/" + results_dir)
    utils.ensure_dir_exists(results_dir)

    args = ["oscap", "xccdf", "eval"]
    args.append("--remediate")
    args.append("--results=%s" % RESULTS_PATH)
    args.append("--report=%s" % REPORT_PATH)

    # oscap uses the default profile by default
    if profile.lower() != "default":
        args.append("--profile=%s" % profile)
    if ds_id:
        args.append("--datastream-id=%s" % ds_id)
    if xccdf_id:
        args.append("--xccdf-id=%s" % xccdf_id)
    if tailoring:
        args.append("--tailoring-file=%s" % tailoring)

    args.append(fpath)

    try:
        proc = subprocess.Popen(args,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                preexec_fn=do_chroot)
    except OSError as oserr:
        msg = "Failed to run the oscap tool: %s" % oserr
        raise OSCAPaddonError(msg)

    (stdout, stderr) = proc.communicate()

    messages = re.findall(r'OpenSCAP Error:.*', stderr)
    if messages:
        for message in messages:
            log.warning("OSCAP addon: " + message)

    # save stdout?
    # pylint thinks Popen has no attribute returncode
    # pylint: disable-msg=E1101
    if proc.returncode not in (0, 2):
        # 0 -- success; 2 -- no error, but checks/remediation failed
        msg = "Content evaluation and remediation with the oscap tool "\
            "failed: %s" % stderr
        raise OSCAPaddonError(msg)

    return stdout


def wait_and_fetch_net_data(url, out_file, ca_certs=None):
    """
    Function that waits for network connection and starts a thread that fetches
    data over network.

    :see: org_fedora_oscap.data_fetch.fetch_data
    :return: the name of the thread running fetch_data
    :rtype: str

    """

    # get thread that tries to establish a network connection
    nm_conn_thread = threadMgr.get(constants.THREAD_WAIT_FOR_CONNECTING_NM)
    if nm_conn_thread:
        # NM still connecting, wait for it to finish
        nm_conn_thread.join()

    if not nm.nm_is_connected():
        raise OSCAPaddonNetworkError("Network connection needed to fetch data.")

    fetch_data_thread = AnacondaThread(name=THREAD_FETCH_DATA,
                                       target=fetch_data,
                                       args=(url, out_file, ca_certs),
                                       fatal=False)

    # register and run the thread
    threadMgr.add(fetch_data_thread)

    return THREAD_FETCH_DATA


def extract_data(archive, out_dir, ensure_has_files=None):
    """
    Fuction that extracts the given archive to the given output directory. It
    tries to find out the archive type by the file name.

    :param archive: path to the archive file that should be extracted
    :type archive: str
    :param out_dir: output directory the archive should be extracted to
    :type out_dir: str
    :param ensure_has_files: relative paths to the files that must exist in the
                             archive
    :type ensure_has_files: iterable of strings or None
    :return: a list of files and directories extracted from the archive
    :rtype: [str]

    """

    # get rid of empty file paths
    ensure_has_files = [fpath for fpath in ensure_has_files if fpath]

    if archive.endswith(".zip"):
        # ZIP file
        try:
            zfile = zipfile.ZipFile(archive, "r")
        except zipfile.BadZipfile as err:
            raise ExtractionError(err.message)

        # generator for the paths of the files found in the archive (dirs end
        # with "/")
        files = set(info.filename for info in zfile.filelist
                    if not info.filename.endswith("/"))
        for fpath in ensure_has_files or ():
            if fpath not in files:
                msg = "File '%s' not found in the archive '%s'" % (fpath,
                                                                   archive)
                raise ExtractionError(msg)

        utils.ensure_dir_exists(out_dir)
        zfile.extractall(path=out_dir)
        zfile.close()
        return [utils.join_paths(out_dir, info.filename) for info in zfile.filelist]
    elif archive.endswith(".tar"):
        # plain tarball
        return _extract_tarball(archive, out_dir, ensure_has_files, None)
    elif archive.endswith(".tar.gz"):
        # gzipped tarball
        return _extract_tarball(archive, out_dir, ensure_has_files, "gz")
    elif archive.endswith(".tar.bz2"):
        # bzipped tarball
        return _extract_tarball(archive, out_dir, ensure_has_files, "bz2")
    elif archive.endswith(".rpm"):
        # RPM
        return _extract_rpm(archive, out_dir, ensure_has_files)
    # elif other types of archives
    else:
        raise ExtractionError("Unsuported archive type")


def _extract_tarball(archive, out_dir, ensure_has_files, alg):
    """
    Extract the given TAR archive to the given output directory and make sure
    the given file exists in the archive.

    :see: extract_data
    :param alg: compression algorithm used for the tarball
    :type alg: str (one of "gz", "bz2") or None
    :return: a list of files and directories extracted from the archive
    :rtype: [str]

    """

    if alg and alg not in ("gz", "bz2",):
        raise ExtractionError("Unsupported compression algorithm")

    mode = "r"
    if alg:
        mode += ":%s" % alg

    try:
        tfile = tarfile.TarFile.open(archive, mode)
    except tarfile.TarError as err:
        raise ExtractionError(err.message)

    # generator for the paths of the files found in the archive
    files = set(member.path for member in tfile.getmembers()
                if member.isfile())

    for fpath in ensure_has_files or ():
        if fpath not in files:
            msg = "File '%s' not found in the archive '%s'" % (fpath, archive)
            raise ExtractionError(msg)

    utils.ensure_dir_exists(out_dir)
    tfile.extractall(path=out_dir)
    tfile.close()

    return [utils.join_paths(out_dir, member.path) for member in tfile.getmembers()]


def _extract_rpm(rpm_path, root="/", ensure_has_files=None):
    """
    Extract the given RPM into the directory tree given by the root argument
    and make sure the given file exists in the archive.

    :param rpm_path: path to the RPM file that should be extracted
    :type rpm_path: str
    :param root: root of the directory tree the RPM should be extracted into
    :type root: str
    :param ensure_has_files: relative paths to the files that must exist in the
                             RPM
    :type ensure_has_files: iterable of strings or None
    :return: a list of files and directories extracted from the archive
    :rtype: [str]

    """

    # run rpm2cpio and process the output with the cpioarchive module
    temp_fd, temp_path = tempfile.mkstemp(prefix="oscap_rpm")
    proc = subprocess.Popen(["rpm2cpio", rpm_path], stdout=temp_fd)
    proc.wait()
    if proc.returncode != 0:
        msg = "Failed to convert RPM '%s' to cpio archive" % rpm_path
        raise ExtractionError(msg)

    os.close(temp_fd)

    try:
        archive = cpioarchive.CpioArchive(temp_path)
    except cpioarchive.CpioError as err:
        raise ExtractionError(err.message)

    # get entries from the archive (supports only iteration over entries)
    entries = set(entry for entry in archive)

    # cpio entry names (paths) start with the dot
    entry_names = [entry.name.lstrip(".") for entry in entries]

    for fpath in ensure_has_files or ():
        # RPM->cpio entries have absolute paths
        if fpath not in entry_names and \
           os.path.join("/", fpath) not in entry_names:
            msg = "File '%s' not found in the archive '%s'" % (fpath, rpm_path)
            raise ExtractionError(msg)

    try:
        for entry in entries:
            dirname = os.path.dirname(entry.name.lstrip("."))
            out_dir = os.path.normpath(root + dirname)
            utils.ensure_dir_exists(out_dir)

            out_fpath = os.path.normpath(root + entry.name.lstrip("."))
            if os.path.exists(out_fpath):
                continue
            with open(out_fpath, "wb") as out_file:
                buf = entry.read(IO_BUF_SIZE)
                while buf:
                    out_file.write(buf)
                    buf = entry.read(IO_BUF_SIZE)
    except (IOError, cpioarchive.CpioError) as e:
        raise ExtractionError(e)

    # cleanup
    archive.close()
    os.unlink(temp_path)

    return [os.path.normpath(root + name) for name in entry_names]


def strip_content_dir(fpaths, phase="preinst"):
    """
    Strip content directory prefix from the file paths for either
    pre-installation or post-installation phase.

    :param fpaths: iterable of file paths to strip content directory prefix
                   from
    :type fpaths: iterable of strings
    :param phase: specifies pre-installation or post-installation phase
    :type phase: "preinst" or "postinst"
    :return: the same iterable of file paths as given with the content
             directory prefix stripped
    :rtype: same type as fpaths

    """

    if phase == "preinst":
        remove_prefix = lambda x: x[len(INSTALLATION_CONTENT_DIR):]
    else:
        remove_prefix = lambda x: x[len(TARGET_CONTENT_DIR):]

    return utils.keep_type_map(remove_prefix, fpaths)


def ssg_available(root="/"):
    """
    Tries to find the SCAP Security Guide under the given root.

    :return: True if SSG was found under the given root, False otherwise

    """

    return os.path.exists(utils.join_paths(root, SSG_DIR + SSG_CONTENT))


def dry_run_skip(func):
    """
    Decorator that makes sure the decorated function is noop in the dry-run
    mode.

    :param func: a decorated function that needs to have the first parameter an
                 object with the _addon_data attribute referencing the OSCAP
                 addon's ksdata
    """

    @wraps(func)
    def decorated(self, *args, **kwargs):
        if self._addon_data.dry_run:
            return
        else:
            return func(self, *args, **kwargs)

    return decorated
